import unittest
from unittest.mock import Mock
from types import ModuleType

from qmi.core.pubsub import QMI_SignalReceiver
from qmi.utils.context_managers import start_stop, start_stop_join, open_close, lock_unlock, subscribe_unsubscribe

from tests.utils import test_module


def reveal_type(oobj, iobj):
    """Assert that the output object from the context manager is of same type as the input."""
    assert isinstance(iobj, oobj)


class StartStop:
    def start(self, *args, **kwargs) -> None: ...
    def stop(self) -> None: ...


class StartStopNoArgs:
    def start(self) -> None: ...
    def stop(self) -> None: ...


class StartStopJoin:
    def start(self, *args, **kwargs) -> None: ...
    def stop(self) -> None: ...
    def join(self) -> None: ...


class SubscribeUnsubscribe:
    def subscribe(self, *args) -> None: ...
    def unsubscribe(self, *args) -> None: ...


class SubscribeUnsubscribeNoArgs:
    def subscribe(self) -> None: ...
    def unsubscribe(self) -> None: ...


class StartOnly:
    def start(self): ...


class OpenOnly:
    def open(self): ...


class LockOnly:
    def lock(self): ...
    
    
class SubscribeOnly:
    def subscribe(self, *args) -> None: ...


class TestContextManagers(unittest.TestCase):
    """Tests the context manager classes."""

    def test_with_start_stop(self) -> None:
        """Test that a right type of class has to be given for a `start_stop` context manager for it to work."""
        iobj_ss = StartStop
        with start_stop(iobj_ss(), 'foo', bar='bar') as res1:
            reveal_type(iobj_ss, res1)  # Check if typed as StartStop

        iobj_ssna = StartStopNoArgs
        with start_stop(iobj_ssna()) as res2:
            reveal_type(iobj_ssna, res2)  # Check if typed as StartStopNoArgs

        with start_stop(test_module, "foo", bar=1) as res3:  # type: ignore
            reveal_type(ModuleType, res3)  # Check if typed as a module

        iobj = Mock
        with start_stop(iobj(), 'foo', bar='bar') as res4:
            reveal_type(iobj, res4)  # Check if typed as Mock

        res4.start.assert_called_once_with('foo', bar='bar')
        res4.stop.assert_called_once()

        # Tests with incomplete or non-correct objects
        exp_err_1 = "'object' object has no attribute 'start'"
        exp_err_2 = "'StartOnly' object has no attribute 'stop'"
        with self.assertRaises(AttributeError) as err1:
            with start_stop(object()):  # type: ignore
                pass

        with self.assertRaises(AttributeError) as err2:
            with start_stop(StartOnly()):  # type: ignore
                pass

        self.assertEqual(exp_err_1, str(err1.exception))
        self.assertEqual(exp_err_2, str(err2.exception))

    def test_with_start_stop_join(self) -> None:
        """Test that a right type of class has to be given for a `start_stop_join` context manager for it to work."""
        iobj_ssj = StartStopJoin
        with start_stop_join(iobj_ssj()) as res1:
            reveal_type(iobj_ssj, res1)  # Check if typed as StartStopJoin

        iobj = Mock
        with start_stop(iobj()) as res2:
            reveal_type(iobj, res2)  # Check if typed as Mock

        with start_stop_join(iobj()) as res3:
            reveal_type(iobj, res3)  # Check if typed as Mock
            pass

        res2.join.assert_not_called()
        res3.join.assert_called()

        # Test with incomplete object
        exp_err_1 = "'StartStopNoArgs' object has no attribute 'join'"
        with self.assertRaises(AttributeError) as err1:
            with start_stop_join(StartStopNoArgs()):  # type: ignore
                pass

        self.assertEqual(exp_err_1, str(err1.exception))

    def test_with_open_close(self) -> None:
        """Test that a right type of class has to be given for a `open_close` context manager for it to work."""
        iobj = Mock
        with open_close(iobj()) as oc:
            reveal_type(iobj, oc)  # Check if types as Mock

        oc.open.assert_called_once()
        oc.close.assert_called_once()

        # Tests with incomplete or non-correct objects
        exp_err_1 = "open_close() got an unexpected keyword argument 'bar'"
        exp_err_2 = "'object' object has no attribute 'open'"
        exp_err_3 = "'OpenOnly' object has no attribute 'close'"
        with self.assertRaises(TypeError) as err1:
            with open_close(iobj(), 'foo', bar='bar'):  # type: ignore
                pass

        with self.assertRaises(AttributeError) as err2:
            with open_close(object()):  # type: ignore
                pass

        with self.assertRaises(AttributeError) as err3:
            with open_close(OpenOnly()):  # type: ignore
                pass

        self.assertEqual(exp_err_1, str(err1.exception))
        self.assertEqual(exp_err_2, str(err2.exception))
        self.assertEqual(exp_err_3, str(err3.exception))

    def test_with_lock_unlock(self) -> None:
        """Test that a right type of class has to be given for a `lock_unlock` context manager for it to work."""
        iobj = Mock
        with lock_unlock(iobj(), 'foo', bar='bar') as lu:
            reveal_type(iobj, lu)  # Check if types as Mock

        lu.lock.assert_called_once_with('foo', bar='bar')
        lu.unlock.assert_called_once()

        # Tests with incomplete or non-correct objects
        exp_error_1 = "'object' object has no attribute 'lock'"
        exp_error_2 = "'LockOnly' object has no attribute 'unlock'"
        with self.assertRaises(AttributeError) as err1:
            with lock_unlock(object()):  # type: ignore
                pass

        with self.assertRaises(AttributeError) as err2:
            with lock_unlock(LockOnly()):  # type: ignore
                pass

        self.assertEqual(exp_error_1, str(err1.exception))
        self.assertEqual(exp_error_2, str(err2.exception))

    def test_with_subscribe_unsubscribe(self) -> None:
        """Test that a right type of class has to be given for a
        `subscribe_unsubscribe` context manager for it to work."""
        iobj = Mock()
        receiver = QMI_SignalReceiver()
        with subscribe_unsubscribe(iobj, receiver) as lu:
            reveal_type(QMI_SignalReceiver, lu)

        iobj.subscribe.assert_called_once_with(receiver)
        iobj.unsubscribe.assert_called_once_with(receiver)

        # It should also work with 'None'
        iobj.reset_mock()
        with subscribe_unsubscribe(iobj, None) as lu:
            reveal_type(QMI_SignalReceiver, lu)

        iobj.subscribe.assert_called_once()
        iobj.unsubscribe.assert_called_once()

        # Tests with incomplete or non-correct objects
        exp_error_1 = "'object' object has no attribute 'subscribe'"
        exp_error_2 = "'SubscribeOnly' object has no attribute 'unsubscribe'"
        exp_error_3 = "subscribe() takes 1 positional argument but 2 were given"

        with self.assertRaises(AttributeError) as err1:
            with subscribe_unsubscribe(object(), None):  # type: ignore
                pass

        with self.assertRaises(AttributeError) as err2:
            with subscribe_unsubscribe(SubscribeOnly(), object()):  # type: ignore
                pass

        with self.assertRaises(TypeError) as err3:
            with subscribe_unsubscribe(SubscribeUnsubscribeNoArgs(), None):
                pass

        self.assertEqual(exp_error_1, str(err1.exception))
        self.assertEqual(exp_error_2, str(err2.exception))
        self.assertTrue(exp_error_3 in str(err3.exception))


if __name__ == '__main__':
    unittest.main()



"""Test for the Boston MicroMachines Multi-DM series deformable mirror QMI driver."""

import logging

from unittest import TestCase
from unittest.mock import MagicMock
from unittest.mock import patch
from unittest.mock import call

from typing import cast
from dataclasses import dataclass

from qmi.instruments.boston_micromachines.multidm import BostonMicromachines_MultiDM
from qmi.instruments.boston_micromachines.multidm import read_mirror_shape
from qmi.instruments.boston_micromachines.multidm import write_mirror_shape
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.exceptions import QMI_UsageException


# Disable all logging
logging.disable(logging.CRITICAL)


@dataclass
class TestMeta:
    """Test meta data."""

    name: MagicMock
    serial_number: MagicMock
    set_shape_prog: MagicMock
    super: MagicMock
    instr: BostonMicromachines_MultiDM


class TestReadMirrorShape(TestCase):
    def test_read_mirror_shape(self):
        """read_mirror_shape(), happy flow"""
        mock_file_name = MagicMock()
        mock_file = MagicMock()
        mock_file.readline = MagicMock(side_effect=["0.1\r\n"] * 3 + [None])
        mock_file.__enter__.return_value = mock_file
        mock_open = MagicMock(return_value=mock_file)
        with patch("qmi.instruments.boston_micromachines.multidm.open", mock_open):
            rt_val = read_mirror_shape(mock_file_name)
        mock_open.assert_called_once_with(mock_file_name, "r")
        self.assertEqual(rt_val, [0.1] * 3)


class TestWriteMirrorShape(TestCase):
    def test_write_mirror_shape(self):
        """write_mirror_shape(), happy flow"""
        mock_file_name = MagicMock()
        mock_shape = [0.1, 0.2, 0.3]
        mock_file = MagicMock()
        mock_file.__enter__.return_value = mock_file
        mock_open = MagicMock(return_value=mock_file)
        mock_print = MagicMock()
        with patch(
            "qmi.instruments.boston_micromachines.multidm.open", mock_open
        ), patch("qmi.instruments.boston_micromachines.multidm.print", mock_print):
            write_mirror_shape(mock_file_name, mock_shape)
        mock_open.assert_called_once_with(mock_file_name, "w")
        mock_print.assert_has_calls(
            [call(f"{i:.9f}", file=mock_file) for i in mock_shape]
        )


class TestBostonMicromachinesMultiDM(TestCase):
    """Testcase for the BostonMicromachines_MultiDM class."""

    def setUp(self):
        mock_name = MagicMock()
        mock_serial_number = "12345678910"
        mock_set_shape_prog = MagicMock()
        mock_super = MagicMock()

        instr = BostonMicromachines_MultiDM(
            MagicMock(), mock_name, mock_serial_number, mock_set_shape_prog
        )

        self._patcher_super = patch(
            "qmi.instruments.boston_micromachines.multidm.super", mock_super
        )
        self._patcher_super.start()

        self._meta = TestMeta(
            instr=cast(BostonMicromachines_MultiDM, instr),
            name=mock_name,
            serial_number=mock_serial_number,
            set_shape_prog=mock_set_shape_prog,
            super=mock_super,
        )

    def tearDown(self):
        self._meta = None
        self._patcher_super.stop()

    def test_init(self):
        """BostonMicromachines_MultiDM.__init__(), happy flow"""

        self.assertEqual(self._meta.instr._serial_number, self._meta.serial_number)
        self.assertEqual(self._meta.instr._set_shape_prog, self._meta.set_shape_prog)

    def test_init_invalid_serial_len(self):
        """BostonMicromachines_MultiDM.__init__(), invalid serial length"""
        mock_serial_number = "1234567891"

        with self.assertRaises(QMI_InstrumentException):
            BostonMicromachines_MultiDM(
                MagicMock(), MagicMock(), mock_serial_number, MagicMock()
            )

    def test_init_invalid_serial_character(self):
        """BostonMicromachines_MultiDM.__init__(), invalid serial character"""
        mock_serial_number = "1234567891!"

        with self.assertRaises(QMI_InstrumentException):
            BostonMicromachines_MultiDM(
                MagicMock(), MagicMock(), mock_serial_number, MagicMock()
            )

    def test_run_set_shape_prog(self):
        """BostonMicromachines_MultiDM._run_set_shape_prog(), happy flow"""
        mock_shape = [0.001, 0.002, 0.003]
        mock_subprocess = MagicMock()
        mock_proc = MagicMock()
        mock_subprocess.Popen = MagicMock(return_value=mock_proc)
        mock_helper_output = "OK\nOK\nOK\n".encode("latin1")
        mock_helper_input = MagicMock()
        mock_proc.communicate = MagicMock(
            return_value=(mock_helper_output, mock_helper_input)
        )
        mock_proc.returncode = 0

        with patch(
            "qmi.instruments.boston_micromachines.multidm.subprocess", mock_subprocess
        ):
            self._meta.instr._run_set_shape_prog(mock_shape)

        mock_subprocess.Popen.assert_called_once_with(
            args=[self._meta.set_shape_prog, "-s", self._meta.serial_number, "-i"],
            stdin=mock_subprocess.PIPE,
            stdout=mock_subprocess.PIPE,
        )
        mock_proc.kill.assert_called_once_with()
        mock_proc.stdin.close.assert_called_once_with()
        mock_proc.wait.assert_called_once_with()
        mock_proc.stdout.close.assert_called_once_with()

    def test_run_set_shape_prog_helper_error(self):
        """BostonMicromachines_MultiDM._run_set_shape_prog(), helper errors handling."""
        mock_shape = [0.001, 0.002, 0.003]
        mock_helper_output = "ERROR\nOK\nOK\n".encode("latin1")

        mock_subprocess = MagicMock()
        mock_proc = MagicMock()
        mock_subprocess.Popen = MagicMock(return_value=mock_proc)
        mock_helper_input = MagicMock()
        mock_proc.communicate = MagicMock(
            return_value=(mock_helper_output, mock_helper_input)
        )
        mock_proc.returncode = 0

        with patch(
            "qmi.instruments.boston_micromachines.multidm.subprocess", mock_subprocess
        ):
            with self.assertRaises(QMI_InstrumentException):
                self._meta.instr._run_set_shape_prog(mock_shape)

        mock_subprocess.Popen.assert_called_once_with(
            args=[self._meta.set_shape_prog, "-s", self._meta.serial_number, "-i"],
            stdin=mock_subprocess.PIPE,
            stdout=mock_subprocess.PIPE,
        )
        mock_proc.kill.assert_called_once_with()
        mock_proc.stdin.close.assert_called_once_with()
        mock_proc.wait.assert_called_once_with()
        mock_proc.stdout.close.assert_called_once_with()

    def test_run_set_shape_prog_communicate_failure(self):
        """BostonMicromachines_MultiDM._run_set_shape_prog(), communicate failure handling."""
        mock_shape = [0.001, 0.002, 0.003]
        mock_helper_output = "OK\nOK\nOK\n".encode("latin1")
        mock_subprocess = MagicMock()
        mock_proc = MagicMock()
        mock_subprocess.Popen = MagicMock(return_value=mock_proc)
        mock_helper_input = MagicMock()
        mock_proc.communicate = MagicMock(
            return_value=(mock_helper_output, mock_helper_input)
        )
        mock_proc.returncode = 1

        with patch(
            "qmi.instruments.boston_micromachines.multidm.subprocess", mock_subprocess
        ):
            with self.assertRaises(QMI_InstrumentException):
                self._meta.instr._run_set_shape_prog(mock_shape)

        mock_subprocess.Popen.assert_called_once_with(
            args=[self._meta.set_shape_prog, "-s", self._meta.serial_number, "-i"],
            stdin=mock_subprocess.PIPE,
            stdout=mock_subprocess.PIPE,
        )
        mock_proc.kill.assert_called_once_with()
        mock_proc.stdin.close.assert_called_once_with()
        mock_proc.wait.assert_called_once_with()
        mock_proc.stdout.close.assert_called_once_with()

    def test_open(self):
        """BostonMicromachines_MultiDM.open(), happy flow."""
        mock_isfile = MagicMock()
        with patch(
            "qmi.instruments.boston_micromachines.multidm.os.path.isfile", mock_isfile
        ):
            self._meta.instr.open()
        self._meta.super().open.assert_called_once_with()
        mock_isfile.assert_called_once_with(self._meta.set_shape_prog)

    def test_open_invalid_path(self):
        """BostonMicromachines_MultiDM.open(), invalid path handling."""
        mock_isfile = MagicMock(return_value=False)
        with patch(
            "qmi.instruments.boston_micromachines.multidm.os.path.isfile", mock_isfile
        ):
            with self.assertRaises(QMI_InstrumentException):
                self._meta.instr.open()
        mock_isfile.assert_called_once_with(self._meta.set_shape_prog)

    def test_apply(self):
        """BostonMicromachines_MultiDM.apply(), happy flow."""
        mock_shape = [0.01] * 140

        self._meta.instr._check_is_open = MagicMock()
        self._meta.instr._run_set_shape_prog = MagicMock()

        self._meta.instr.apply(mock_shape)

        self._meta.instr._check_is_open.assert_called_once_with()
        self._meta.instr._run_set_shape_prog.assert_called_once_with(mock_shape)

    def test_apply_invalid_shape_len(self):
        """BostonMicromachines_MultiDM.apply(), invalid shape len handling."""
        mock_shape = []

        self._meta.instr._check_is_open = MagicMock()
        with self.assertRaises(QMI_UsageException):
            self._meta.instr.apply(mock_shape)

        self._meta.instr._check_is_open.assert_called_once_with()

    def test_apply_value_to_high(self):
        """BostonMicromachines_MultiDM.apply(), value too high."""
        mock_shape = [0.01] * 139 + [1.1]

        self._meta.instr._check_is_open = MagicMock()
        with self.assertRaises(QMI_UsageException):
            self._meta.instr.apply(mock_shape)

        self._meta.instr._check_is_open.assert_called_once_with()

    def test_apply_value_to_low(self):
        """BostonMicromachines_MultiDM.apply(), value too low."""
        mock_shape = [0.01] * 139 + [-0.1]

        self._meta.instr._check_is_open = MagicMock()
        with self.assertRaises(QMI_UsageException):
            self._meta.instr.apply(mock_shape)

        self._meta.instr._check_is_open.assert_called_once_with()



#! /usr/bin/env python3

"""Test configuration file processing."""

import unittest
import json
import os
import tempfile

import qmi.core.config
import qmi.core.exceptions


CONFIG_BASIC = """# TEST FILE FOR CONFIG PARSING
{
    "name": "config",
    "num": 13,
    "pi": 3.1415, # this is an approximation
    "nocomment": "the # sign is not special inside strings",
    "comment": #" but this is a comment",
        "and this is the value",
    #"skipped": 18,
    "#": "pound"
}"""

CONFIG_DUPLICATE = """# TEST FILE FOR CONFIG PARSING
{
    "name": "config",
    "num": 13,
    "pi": 3.1415, # this is an approximation
    "nocomment": "the # sign is not special inside strings",
    "comment": #" but this is a comment",
        "and this is the value",
    "pi": 3.1415, # this is a duplicate
    #"skipped": 18,
    "#": "pound"
}"""

PARSED_BASIC = {
    "name": "config",
    "num": 13,
    "pi": 3.1415,
    "nocomment": "the # sign is not special inside strings",
    "comment": "and this is the value",
    "#": "pound"
}

CONFIG_TRICKY = r"""{#}
    "nocomment1": "\"#\"",
    "nocomment2": "\\#",
    "nocomment3": "\\\"#",
    "comment1": "\\",#",
    "comment2": "\u00c0\n"#',
  , "comment3":# "\uuuuu",
        1,
    "comment4"#: "fake",
        :"real",
    "comment5": 5#.0
  , "comment6": [#]
        6, "6"],
    "comment7": #null,
        7,
    "comment8": true#false,
        ,
    "comment9": {#},
        },
    "comment10": { "name"#} : "fake",
        :#"fake" }
"value"##,
}#}}
,
#   "commentX":
    "#nocomment4": 4,
    "nocomment5#": 5,
    "nocomment6":"#6",
    "nocomment7": "7#"#,
  ,#null
    "nocomment8\"#": 333,
# "#" \# \# \#
    "array": [[]#],
        ],
    "null": null
#}}}
}#
"""

PARSED_TRICKY = {
    "nocomment1": '"#"',
    "nocomment2": "\\#",
    "nocomment3": '\\"#',
    "comment1": "\\",
    "comment2": "\xc0\n",
    "comment3": 1,
    "comment4": "real",
    "comment5": 5,
    "comment6": [6, "6"],
    "comment7": 7,
    "comment8": True,
    "comment9": {},
    "comment10": { "name": "value" },
    "#nocomment4": 4,
    "nocomment5#": 5,
    "nocomment6": "#6",
    "nocomment7": "7#",
    'nocomment8"#': 333,
    "array": [[]],
    "null": None
}

# Invalid config: Bad JSON string.
CONFIG_BAD_JSON = """{ "bad": False }"""

# Invalid config: top-level type must be a map.
CONFIG_BAD_TYPE = """[ 3, 4, 5 ]"""


class TestConfigLoad(unittest.TestCase):

    def test_basic(self):
        cfg = qmi.core.config.load_config_string(CONFIG_BASIC)
        self.assertEqual(cfg, PARSED_BASIC)

    def test_tricky(self):
        cfg = qmi.core.config.load_config_string(CONFIG_TRICKY)
        self.assertEqual(cfg, PARSED_TRICKY)

    def test_duplicate(self):
        with self.assertRaises(ValueError):
            cfg = qmi.core.config.load_config_string(CONFIG_DUPLICATE)

    def test_bad_json(self):
        with self.assertRaises(json.JSONDecodeError):
            cfg = qmi.core.config.load_config_string(CONFIG_BAD_JSON)

    def test_bad_type(self):
        with self.assertRaises(qmi.core.exceptions.QMI_ConfigurationException):
            cfg = qmi.core.config.load_config_string(CONFIG_BAD_TYPE)

    def test_file(self):
        with tempfile.TemporaryDirectory() as tmpdirname:
            filename = os.path.join(tmpdirname, "test_load_basic.cfg")
            with open(filename, "w") as f:
                f.write(CONFIG_BASIC)
            cfg = qmi.core.config.load_config_file(filename)
        self.assertEqual(cfg, PARSED_BASIC)


class TestConfigDump(unittest.TestCase):

    def test_basic(self):
        s = qmi.core.config.dump_config_string(PARSED_BASIC)
        cfg = json.loads(s)
        self.assertEqual(cfg, PARSED_BASIC)

    def test_tricky(self):
        s = qmi.core.config.dump_config_string(PARSED_TRICKY)
        cfg = json.loads(s)
        self.assertEqual(cfg, PARSED_TRICKY)

    def test_file(self):
        with tempfile.TemporaryDirectory() as tmpdirname:
            filename = os.path.join(tmpdirname, "test_dump_basic.cfg")
            qmi.core.config.dump_config_file(PARSED_BASIC, filename)
            with open(filename, "r") as f:
                s = f.read()
        cfg = json.loads(s)
        self.assertEqual(cfg, PARSED_BASIC)

    def test_file_overwrite(self):
        with tempfile.TemporaryDirectory() as tmpdirname:
            filename = os.path.join(tmpdirname, "test_dump_overwrite.cfg")
            with open(filename, "w") as f:
                f.write("dummy")
            qmi.core.config.dump_config_file(PARSED_BASIC, filename)
            with open(filename, "r") as f:
                s = f.read()
        cfg = json.loads(s)
        self.assertEqual(cfg, PARSED_BASIC)


if __name__ == "__main__":
    unittest.main()



"""Unit-tests for Teraxion TFN."""
from datetime import datetime
import unittest
from unittest.mock import call, patch

from qmi.core.context import QMI_Context
from qmi.instruments.teraxion.tfn import Teraxion_TFN, Teraxion_TFNChannelPlan, Teraxion_TFNElement, Teraxion_TFNSettings, Teraxion_TFNStatus


class TestTeraxionTfn(unittest.TestCase):

    def setUp(self):
        qmi_context = unittest.mock.MagicMock(spec=QMI_Context)
        qmi_context.name = "mock_context"

        # Add patches
        with patch("qmi.instruments.teraxion.tfn.create_transport") as self._transport_mock:
            self.tfn = Teraxion_TFN(qmi_context, "teraxion_tfn", "")
        self.tfn.open()

    def tearDown(self) -> None:
        self.tfn.close()

    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.ask")
    def test_get_firmware_version_gets_firmware_version(self, ask_mock):
        """Test get firmware version, gets firmware version."""
        # Arrange
        expected_version = "1.0"
        expected_command = "S600fP S6106P"
        ask_mock.return_value = "001000000100"

        # Act
        ver = self.tfn.get_firmware_version()

        # Assert
        ask_mock.assert_called_once_with(expected_command, timeout=self.tfn.DEFAULT_READ_TIMEOUT)
        self.assertEqual(ver, expected_version)

    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.ask")
    def test_get_manufacturer_name_gets_manufacturer_name(self, ask_mock):
        """Test get manufacturer name, gets manufacturer name."""
        # Arrange
        expected_manufacturer_name = "TeraXion"
        expected_command = "S600eP S61ffP"
        ask_mock.return_value = "001000005465726158696F6E00"

        # Act
        manufacturer_name = self.tfn.get_manufacturer_name()

        # Assert
        ask_mock.assert_called_once_with(expected_command, timeout=self.tfn.DEFAULT_READ_TIMEOUT)
        self.assertEqual(manufacturer_name, expected_manufacturer_name)

    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.ask")
    def test_get_model_number_gets_model_number(self, ask_mock):
        """Test get model number, gets model number."""
        # Arrange
        expected_model_number = "TFN-XXXX"
        expected_command = "S6027P S61ffP"
        ask_mock.return_value = "0010000054464E2D5858585800"

        # Act
        model_number = self.tfn.get_model_number()

        # Assert
        ask_mock.assert_called_once_with(expected_command, timeout=self.tfn.DEFAULT_READ_TIMEOUT)
        self.assertEqual(model_number, expected_model_number)

    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.ask")
    def test_get_serial_number_gets_serial_number(self, ask_mock):
        """Test get serial number, gets serial number."""
        # Arrange
        expected_serial_number = "Txxxxxx"
        expected_command = "S6029P S61ffP"
        ask_mock.return_value = "001000005478787878787800"

        # Act
        serial_number = self.tfn.get_serial_number()

        # Assert
        ask_mock.assert_called_once_with(expected_command, timeout=self.tfn.DEFAULT_READ_TIMEOUT)
        self.assertEqual(serial_number, expected_serial_number)

    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.ask")
    def test_get_idn_sends_get_idn_command(self, ask_mock):
        """Test get idn, sends gets idn command."""
        # Arrange
        expected_vendor = "TeraXion"
        expected_model_number = "TFN-XXXX"
        expected_serial_number = "Txxxxxx"
        expected_version = "1.0"
        expected_manufactuer_name_command = "S600eP S61ffP"
        expected_model_number_command = "S6027P S61ffP"
        expected_serial_number_command = "S6029P S61ffP"
        expected_firmware_version_command = "S600fP S6106P"

        ask_mock.side_effect = ["001000005465726158696F6E00", "0010000054464E2D5858585800", "001000005478787878787800", "001000000100"]

        # Act
        idn = self.tfn.get_idn()

        # Assert
        ask_mock.assert_has_calls([call(expected_manufactuer_name_command, timeout=self.tfn.DEFAULT_READ_TIMEOUT),
                                  call(expected_model_number_command, timeout=self.tfn.DEFAULT_READ_TIMEOUT),
                                  call(expected_serial_number_command, timeout=self.tfn.DEFAULT_READ_TIMEOUT),
                                  call(expected_firmware_version_command, timeout=self.tfn.DEFAULT_READ_TIMEOUT)])
        self.assertEqual(idn.vendor, expected_vendor)
        self.assertEqual(idn.model, expected_model_number)
        self.assertEqual(idn.serial, expected_serial_number)
        self.assertEqual(idn.version, expected_version)

    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.ask")
    def test_get_manufacturing_date_gets_manufacturing_date(self, ask_mock):
        """Test get manufacturing number, gets manufacturing date."""
        # Arrange
        expected_date = datetime.strptime("20231120", "%Y%m%d").date()
        expected_command = "S602bP S61ffP"
        ask_mock.return_value = "00100000323032333131323000"

        # Act
        manufacturing_date = self.tfn.get_manufacturing_date()

        # Assert
        ask_mock.assert_called_once_with(expected_command, timeout=self.tfn.DEFAULT_READ_TIMEOUT)
        self.assertEqual(manufacturing_date, expected_date)

    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.ask")
    def test_get_status_gets_status(self, ask_mock):
        """Test get status, gets status."""
        # Arrange
        expected_status = Teraxion_TFNStatus(False, False, False, True, False, False, False, False, False, False, False, False, False, False)
        expected_command = "S6000P S6104P"
        ask_mock.return_value = "00100000"

        # Act
        status = self.tfn.get_status()

        # Assert
        ask_mock.assert_called_once_with(expected_command, timeout=self.tfn.DEFAULT_READ_TIMEOUT)
        self.assertEqual(status, expected_status)

    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.write")
    def test_reset_resets(self, write_mock):
        """Test reset, resets."""
        # Arrange
        expected_command = "S6028P"

        # Act
        self.tfn.reset()

        # Assert
        write_mock.assert_called_once_with(expected_command)

    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.ask")
    def test_set_frequency_sets_frequency(self, ask_mock):
        """Test set frequency, sets frequency."""
        # Arrange
        set_freq = 192400
        expected_command = "S602e483be400P L000a S6108P"

        # Act
        self.tfn.set_frequency(set_freq)

        # Assert
        ask_mock.assert_called_once_with(expected_command, timeout=self.tfn.DEFAULT_READ_TIMEOUT)

    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.ask")
    def test_get_frequency_gets_frequency(self, ask_mock):
        """Test get frequency, gets frequency."""
        # Arrange
        expected_freq = 192400
        expected_command = "S602fP S6108P"
        ask_mock.return_value = "00100000483be400"

        # Act
        freq = self.tfn.get_frequency()

        # Assert
        ask_mock.assert_called_once_with(expected_command, timeout=self.tfn.DEFAULT_READ_TIMEOUT)
        self.assertEqual(freq, expected_freq)

    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.ask")
    def test_get_rtd_temp_given_element_gets_rtd_temp(self, ask_mock):
        """Test get RTD temperature given the element, gets the RTD temperature."""
        # Arrange
        expected_temp = 8918
        expected_command = "S601704P S6106P"
        ask_mock.return_value = "0010000022D6"

        # Act
        temp = self.tfn.get_rtd_temperature(Teraxion_TFNElement.CASE)

        # Assert
        ask_mock.assert_called_once_with(expected_command, timeout=self.tfn.DEFAULT_READ_TIMEOUT)
        self.assertEqual(temp, expected_temp)

    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.ask")
    def test_enable_device_sends_enable_device_command(self, ask_mock):
        """Test enable device, sends enable device command."""
        # Arrange
        expected_command = "S601eP L000a S6104P"

        # Act
        self.tfn.enable_device()

        # Assert
        ask_mock.assert_called_once_with(expected_command, timeout=self.tfn.DEFAULT_READ_TIMEOUT)

    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.ask")
    def test_disable_device_sends_disable_device_command(self, ask_mock):
        """Test disable device, sends disable device command."""
        # Arrange
        expected_command = "S601fP L000a S6104P"

        # Act
        self.tfn.disable_device()

        # Assert
        ask_mock.assert_called_once_with(expected_command, timeout=self.tfn.DEFAULT_READ_TIMEOUT)

    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.ask")
    def test_enable_tecs_on_startup_sends_set_startup_byte_command(self, ask_mock):
        """Test enable tecs on startup, sends startup byte command."""
        # Arrange
        expected_command = "S603401P L000a S6105P"

        # Act
        self.tfn.enable_tecs_on_startup()

        # Assert
        ask_mock.assert_called_once_with(expected_command, timeout=self.tfn.DEFAULT_READ_TIMEOUT)

    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.ask")
    def test_disable_tecs_on_startup_sends_set_startup_byte_command(self, ask_mock):
        """Test disable tecs on startup, sends startup byte command."""
        # Arrange
        expected_command = "S603400P L000a S6105P"

        # Act
        self.tfn.disable_tecs_on_startup()

        # Assert
        ask_mock.assert_called_once_with(expected_command, timeout=self.tfn.DEFAULT_READ_TIMEOUT)

    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.ask")
    def test_get_startup_byte_sends_get_startup_byte_command(self, ask_mock):
        """Test get startup byte, gets startup byte."""
        # Arrange
        expected_command = "S6035P S6105P"
        ask_mock.return_value = "0010000001"

        # Act
        startup_byte = self.tfn.get_startup_byte()

        # Assert
        ask_mock.assert_called_once_with(expected_command, timeout=self.tfn.DEFAULT_READ_TIMEOUT)
        self.assertTrue(startup_byte)

    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.ask")
    def test_get_nominal_settings_gets_nominal_settings(self, ask_mock):
        """Test get nominal settings, gets nominal settings."""
        # Arrange
        expected_settings = Teraxion_TFNSettings(192400, 0)
        expected_command = "S6037P S610cP"
        ask_mock.return_value = "00100000483be40000000000"

        # Act
        settings = self.tfn.get_nominal_settings()

        # Assert
        ask_mock.assert_called_once_with(expected_command, timeout=self.tfn.DEFAULT_READ_TIMEOUT)
        self.assertEqual(settings, expected_settings)

    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.ask")
    def test_save_nominal_settings_sends_save_nominal_settings_command(self, ask_mock):
        """Test save nominal settings, saves nominal settings."""
        # Arrange
        expected_command = "S6036P L000a S610cP"

        # Act
        self.tfn.save_nominal_settings()

        # Assert
        ask_mock.assert_called_once_with(expected_command, timeout=self.tfn.DEFAULT_READ_TIMEOUT)

    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.ask")
    def test_get_channel_plan_gets_channel_plan(self, ask_mock):
        """Test get channel plan, gets channel plan."""
        # Arrange
        expected_channel_plan = Teraxion_TFNChannelPlan(193420, 193370, 1)
        expected_command = "S603bP S6110P"
        ask_mock.return_value = "00100000483CE300483CD68000000001"

        # Act
        channel_plan = self.tfn.get_channel_plan()

        # Assert
        ask_mock.assert_called_once_with(expected_command, timeout=self.tfn.DEFAULT_READ_TIMEOUT)
        self.assertEqual(channel_plan, expected_channel_plan)

    @unittest.mock.patch("qmi.core.scpi_protocol.ScpiProtocol.write")
    def test_set_i2c_address_sends_set_i2c_address_command(self, write_mock):
        """Test set i2c address, sends set i2c address command."""
        # Arrange
        expected_command = "S60420060P"

        # Act
        self.tfn.set_i2c_address(96)

        # Assert
        write_mock.assert_called_once_with(expected_command)


"""
AimTTi: Function generators and frequency counters.

The qmi.instruments.tt package provides support for:
- TGF3000 series (discontinued)
- TGF4000 series.
"""
from qmi.instruments.tt.tgf import FrequencyMeasurement, WaveformType, CounterInputChannel
# Alternative, QMI naming convention approved names
from qmi.instruments.tt.tgf import TT_TGF_3000_4000_Series as AimTTi_Tgf30004000



"""Instrument driver for the Montana Cryostation."""

import logging
import threading
import time
from typing import Optional

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException, QMI_InvalidOperationException, QMI_TimeoutException
from qmi.core.instrument import QMI_Instrument
from qmi.core.rpc import rpc_method
from qmi.core.transport import create_transport

# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


class Montana_Cryostation(QMI_Instrument):
    """Instrument driver for the Montana Cryostation.

    The Montana Cryostation is a cryostat that is delivered with a Windows 10 laptop. On the
    laptop, a GUI application is running. The GUI application allows remote control via TCP
    (by default, port 7773).

    IMPORTANT: the TCP server on the laptop's GUI needs to be manually enabled. There is,
      unfortunately, not a setting to do this automatically on startup.

    The Cryostation is monitored and controlled by a simple ASCII-based protocol as described in [1].

    The documentation lists 29 commands. Of these, 17 retrieve setpoints and sensor values;
    12 are used to alter setpoints or start behavior (e.g. cooldown):

    Getters:

        GAS   - Get Alarm State
        GCP   - Get Chamber Pressure
        GMS   - Get Magnet State
        GMTF  - Get Magnet Target Field
        GPHP  - Get Platform Heater Power
        GPIDF - Get PID Fi integral frequency
        GPIDK - Get PID Ki proportional gain
        GPIDT - Get PID Td derivative time
        GPS   - Get Platform Stability
        GS1T  - Get Stage 1 Temperature
        GS2T  - Get Stage 2 Temperature
        GPT   - Get Platform Temperature
        GSS   - Get Sample Stability
        GST   - Get Sample Temperature
        GTSP  - Get Temperature Set Point
        GUS   - Get User Stability
        GUT   - Get User Temperature
        RPID  - Reset PID parameters to their default values

    Setters:

        SCD   - Start Cool Down
        SMD   - Set Magnet Disabled
        SME   - Set Magnet Enabled
        SMTF  - Set Magnet Target Field
        SPIDF - Set PIDF Fi integral frequency
        SPIDK - Set PID proportional gain
        SPIDT - Set PID Td derivative time
        SSB   - Start StandBy
        STP   - SToP
        STSP  - Set Temperature Set Point
        SWU   - Start Warm Up
        SPHP  - Set Platform Heater Power

    We note that the three commands GPIDF, GPIDK, GPIDT that are used to retrieve PID control parameters
    do not work on our CryoStation.

    Note: this driver assumes that the Montana software uses a locale with '.' as decimal separator (e.g. "en_US"). The
    Montana software inherits this setting from the language settings of the operating system on the monitoring laptop.

    References:

    [1] Montana Cryostation Communication Specification, version 1.9
    """
    # Value ranges
    MIN_PROPORTIONAL_GAIN = 0.001
    MAX_PROPORTIONAL_GAIN = 100.0
    MIN_INTEGRAL_FREQUENCY = 0.0
    MAX_INTEGRAL_FREQUENCY = 100.0
    MIN_DERIVATIVE_TIME = 0.0
    MAX_DERIVATIVE_TIME = 100.0
    MIN_TEMPERATURE_SETPOINT = 2.0
    MAX_TEMPERATURE_SETPOINT = 350.0

    # The Cryostation should respond within 5 seconds.
    RESPONSE_TIMEOUT = 5.0

    # The user should not apply more than 1 Watt to the platform heater.
    _MAX_PLATFORM_HEATER_POWER = 1.0

    # There is a way to set the platform heater up to 10 Watt, but only for a limited duration.
    _MAX_PLATFORM_HEATER_BURST_POWER = 10.0
    _MAX_PLATFORM_HEATER_BURST_DURATION = 30.0

    # Retry turning off the platform heater if necessary.
    _MAX_TURN_OFF_HEATER_RETRY = 3

    def __init__(self, context: QMI_Context, name: str, transport: str) -> None:
        """Initializes a Montana_Cryostation instance.

        Note:
            This method should not be called directly by a QMI user.
            Always instantiate an instrument via a QMI_Context `make_instrument()` call.

        Arguments:
            context: the QMI context
            name: the name of the instrument instance. Use the same name as the variable to which
                  the newly created instance is assigned.
            transport: the transport to the Cryostation software. This should be a TCP transport.
                  The default TCP port number for the Cryostation remote control is 7773.
        """

        super().__init__(context, name)
        self._transport = create_transport(transport, default_attributes={"port": 7773})
        self._timeout = Montana_Cryostation.RESPONSE_TIMEOUT
        self._platform_heater_on = False
        self._platform_heater_timer_thread: Optional[threading.Timer] = None
        self._command_lock = threading.Lock()

    def release_rpc_object(self) -> None:
        """Make sure the platform heater is safely turned off before destroying this instrument driver instance.

        The platform heater is normally turned off by the `close()` method. This function is only necessary
        to handle cases where the driver shuts down without properly calling `close()`.
        """
        super().release_rpc_object()
        if self.is_open():
            self._cancel_platform_heater_burst_timer()
            self._turn_off_platform_heater()

    @rpc_method
    def open(self) -> None:
        """Opens a connection to the Cryostation instrument."""
        _logger.info("Opening connection to %s", self._name)
        self._transport.open()
        super().open()

    @rpc_method
    def close(self) -> None:
        """Closes a connection to the Cryostation instrument."""
        self._check_is_open()
        _logger.info("Closing connection to %s", self._name)

        # Make sure the platform heater is safely turned off before closing the connection.
        self._cancel_platform_heater_burst_timer()
        self._turn_off_platform_heater()

        super().close()
        self._transport.close()

    def _execute_command(self, command: str, *, check_ok=False) -> str:
        """Execute Cryostation command and return response.

        Cryostat commands and responses are ASCII commands preceded by a
        length-of-message field consisting of 2 ASCII decimal digits.
        No end-of-command terminators are used.

        If the `check_ok` flag is set, an extra check is done to see if the 'OK'
        string is present in the response.

        This function can be safely called concurrently from multiple threads.
        We use this to implement automatic shutdown of the platform heater
        via a background timer thread.
        """

        # Use a lock to make sure that only one thread talks to the instrument.
        with self._command_lock:

            # Discard any pending response that was meant for a previous command.
            self._transport.discard_read()

            command_bin = "{:02d}{}".format(len(command), command).encode('ascii')
            self._transport.write(command_bin)
            num_bytes = int(self._transport.read(2, self._timeout))
            response = self._transport.read(num_bytes, self._timeout).decode("ascii")

        if response == "Error: Invalid command.":
            raise QMI_InstrumentException("Cryostat does not recognise command {!r}".format(command))

        if check_ok and "OK" not in response:
            raise QMI_InstrumentException(f"Unexpected response received: {response}!")

        return response

    @rpc_method
    def get_alarm_state(self) -> bool:
        """Retrieves the alarm state, i.e., whether a system error is presently active."""

        command = "GAS"
        response = self._execute_command(command)

        if response not in ["T", "F"]:
            raise ValueError("Bad response from the Cryostat to the {!r} command: {!r}".format(command, response))

        alarm_state = (response == "T")

        return alarm_state

    @rpc_method
    def get_chamber_pressure(self, unit: Optional[str]=None) -> float:
        """Retrieve the current chamber pressure.

        The Cryostation always returns the pressure in [mTorr]. Since that is usually a pretty inconvenient
        unit, this method allows the specification of different units ([Pa], [atm]) that are usually more
        convenient, converting the retrieved value in [mTorr] to that unit is requested.

        Parameters:
            unit: the physical unit of pressure to return. This can be one of 'mTorr', 'Pa', or 'atm'.
                  The default if no unit is specified is mTorr.

        Returns: current chamber pressure in the unit selected, or NaN if the chamber pressure is not available.
        """

        if unit is None:
            unit = 'mTorr'

        if unit not in ['mTorr', 'Pa', 'atm']:
            raise ValueError(f"Bad unit specified for Montana Cryostation get_chamber_pressure method: {unit!r}")

        command = "GCP"
        response = self._execute_command(command)
        pressure = float(response)

        if unit == 'mTorr':
            pass
        elif unit == 'Pa':
            pressure = pressure * (101325.0 / 760000.0)
        elif unit == 'atm':
            pressure = pressure / 760000.0

        return pressure

    @rpc_method
    def get_magnet_state(self) -> Optional[bool]:
        """Retrieves the current magnet state.

        Returns: True if magnet is enabled; False if magnet is disabled;
            None if the system cannot execute the command (activate the magnet first).
        """

        command = "GMS"
        response = self._execute_command(command)

        # "System not able to execute command at this time.  Activate the magnet module first."
        magnet_state: Optional[bool] = None

        if response == "MAGNET ENABLED":
            magnet_state = True
        elif response == "MAGNET DISABLED":
            magnet_state = False

        return magnet_state

    @rpc_method
    def get_magnet_target_field(self) -> float:
        """Retrieves the current set point for magnetic field.

        Returns: current set point for magnetic field in [Tesla], or NaN if the magnet is not enabled or the magnet
            module is not activated.
        """

        command = "GMTF"
        response = self._execute_command(command)

        if response == "-9.999999":
            # The instrument will return -9.999999 if the magnet is not enabled or the magnet module is not activated.
            magnet_target_field = float("nan")
        else:
            magnet_target_field = float(response)

        return magnet_target_field

    @rpc_method
    def get_platform_heater_power(self) -> float:
        """Retrieves the current platform heater power.

        Returns: platform heater power in [Watt], or NaN if the platform heater power is not available.
        """

        command = "GPHP"
        response = self._execute_command(command)

        if response == "-0.100":
            # The instrument will return -0.100 if the platform heater power is not available.
            platform_heater_power = float("nan")
        else:
            platform_heater_power = float(response)

        return platform_heater_power

    @rpc_method
    def get_pid_integral_frequency(self) -> float:
        """Retrieves the current setting for PID Fi integral frequency.

        Returns: PID Fi integral frequency in [Hz].

        Warning: this function doesn't work on our Cryostation and raises a QMI_InstrumentException.
        """

        command = "GPIDF"
        response = self._execute_command(command)

        pid_integral_frequency = float(response)

        return pid_integral_frequency

    @rpc_method
    def set_pid_integral_frequency(self, i_freq_hz: float) -> None:
        """Set the PID integral frequency value.

        Note: 0 will result in no integral action.

        Parameters:
            i_freq: integral frequency in Hertz. (0.000 - 100.000)
        """
        if not self.MIN_INTEGRAL_FREQUENCY <= i_freq_hz <= self.MAX_INTEGRAL_FREQUENCY:
            raise ValueError(f"Integral frequency ({i_freq_hz}) is not in range {(self.MIN_INTEGRAL_FREQUENCY, self.MAX_INTEGRAL_FREQUENCY)}!")

        _logger.info("[%s] Setting integral frequency to %f Hertz", self._name, i_freq_hz)
        self._execute_command(f"SPIDF {i_freq_hz:.3f}", check_ok=True)

    @rpc_method
    def get_pid_proportional_gain(self) -> float:
        """Retrieves the current setting for PID Ki proportional gain.

        Returns: PID Ki proportional gain in [Watt/Kelvin].

        Warning: this function doesn't work on our Cryostation and raises a QMI_InstrumentException.
        """

        command = "GPIDK"
        response = self._execute_command(command)

        pid_proportional_gain = float(response)

        return pid_proportional_gain

    @rpc_method
    def set_pid_proportional_gain(self, p_gain: float) -> None:
        """Set the PID proportional gain value.

        Note: 0 will result in no derivative action.

        Parameters:
            p_gain: Proportional gain in Watts/Kelvin. (0.001 - 100.000)
        """
        if not self.MIN_PROPORTIONAL_GAIN <= p_gain <= self.MAX_PROPORTIONAL_GAIN:
            raise ValueError(f"Proportionial gain ({p_gain}) is not in range {(self.MIN_PROPORTIONAL_GAIN, self.MAX_PROPORTIONAL_GAIN)}!")

        _logger.info("[%s] Setting proportional gain to %f Watts/Kelvin", self._name, p_gain)
        self._execute_command(f"SPIDK {p_gain:.3f}", check_ok=True)

    @rpc_method
    def get_pid_derivative_time(self) -> float:
        """Retrieves the current setting for PID Td derivative time.

        Returns: PID Td derivative time in [seconds].

        Warning: this function doesn't work on our Cryostation and raises a QMI_InstrumentException.
        """

        command = "GPIDT"
        response = self._execute_command(command)

        pid_derivative_time = float(response)

        return pid_derivative_time

    @rpc_method
    def set_pid_derivative_time(self, d_time_s: float) -> None:
        """Set the PID derivative time value.

        Parameters:
            d_time: derivative time in seconds. (0.000 - 100.000)
        """
        if not self.MIN_DERIVATIVE_TIME <= d_time_s <= self.MAX_DERIVATIVE_TIME:
            raise ValueError(f"derivative time ({d_time_s}) is not in range {(self.MIN_DERIVATIVE_TIME, self.MAX_DERIVATIVE_TIME)}!")

        _logger.info("[%s] Setting derivative time to %f seconds", self._name, d_time_s)
        self._execute_command(f"SPIDT {d_time_s:.3f}", check_ok=True)

    @rpc_method
    def get_platform_stability(self) -> float:
        """Retrieves the current platform stability.

        Returns: platform stability in [Kelvin].
        """

        command = "GPS"
        response = self._execute_command(command)

        if response == "-0.10000":
            # The instrument will return -0.10000 if the platform stability is not available.
            platform_stability = float("nan")
        else:
            platform_stability = float(response)

        return platform_stability

    @rpc_method
    def get_stage_1_temperature(self) -> float:
        """Retrieves the current stage 1 temperature.

        Returns: stage 1 temperature in [Kelvin], or NaN if the stage 1 temperature is not available.
        """

        command = "GS1T"
        response = self._execute_command(command)

        if response == "-0.10":
            # The instrument will return -0.10 if the stage 1 temperature is not available.
            stage_1_temperature = float("nan")
        else:
            stage_1_temperature = float(response)

        return stage_1_temperature

    @rpc_method
    def get_stage_2_temperature(self) -> float:
        """Retrieves the current stage 2 temperature.

        Returns: stage 2 temperature in [Kelvin], or NaN if the stage 2 temperature is not available.
        """

        command = "GS2T"
        response = self._execute_command(command)

        if response == "-0.10":
            # The instrument will return -0.10 if the stage 2 temperature is not available.
            stage_2_temperature = float("nan")
        else:
            stage_2_temperature = float(response)

        return stage_2_temperature

    @rpc_method
    def get_platform_temperature(self) -> float:
        """Retrieves the current platform temperature.

        Returns: platform temperature in [Kelvin], or NaN if the platform temperature is not available.
        """

        command = "GPT"
        response = self._execute_command(command)

        if response == "-0.100":
            # The instrument will return -0.100 if the platform temperature is not available.
            temperature = float("nan")
        else:
            temperature = float(response)

        return temperature

    @rpc_method
    def get_sample_stability(self) -> float:
        """Retrieves the current sample stability.

        Returns: sample stability in [Kelvin], or NaN if the sample stability is not available.
        """

        command = "GSS"
        response = self._execute_command(command)

        if response == "-0.10000":
            # The instrument will return -0.10000 if the sample stability is not available.
            sample_stability = float("nan")
        else:
            sample_stability = float(response)

        return sample_stability

    @rpc_method
    def get_sample_temperature(self) -> float:
        """Retrieves the current sample temperature.

        Returns: sample temperature in [Kelvin], or NaN if the sample temperature is not available.
        """

        command = "GST"
        response = self._execute_command(command)

        if response == "-0.100":
            # The instrument will return -0.100 if the sample temperature is not available.
            sample_temperature = float("nan")
        else:
            sample_temperature = float(response)

        return sample_temperature

    @rpc_method
    def get_temperature_setpoint(self) -> float:
        """Retrieves the current temperature setpoint.

        Returns: temperature setpoint in [Kelvin], or NaN if the temperature setpoint is not available.
        """

        command = "GTSP"
        response = self._execute_command(command)

        temperature_setpoint = float(response)

        return temperature_setpoint

    @rpc_method
    def set_temperature_setpoint(self, setpoint_k: float) -> None:
        """Set temperature setpoint.

        Parameters:
            setpoint_k:   Temperature setpoint in Kelvin. (2.00 - 350.00)
        """
        if not self.MIN_TEMPERATURE_SETPOINT <= setpoint_k <= self.MAX_TEMPERATURE_SETPOINT:
            raise ValueError(f"Setpoint ({setpoint_k}) is not in range {(self.MIN_TEMPERATURE_SETPOINT, self.MAX_TEMPERATURE_SETPOINT)}!")

        _logger.info("[%s] Setting temperature setpoint to %f K", self._name, setpoint_k)
        self._execute_command(f"STSP {setpoint_k:.2f}", check_ok=True)

    @rpc_method
    def get_user_stability(self) -> float:
        """Retrieves the current user stability (whatever that may be).

        Returns: user stability in [Kelvin], or NaN if the user stability is not available.
        """

        command = "GUS"
        response = self._execute_command(command)

        if response == "-0.10000":
            # The instrument will return -0.10000 if the user stability is not available.
            user_stability = float("nan")
        else:
            user_stability = float(response)

        return user_stability

    @rpc_method
    def get_user_temperature(self) -> float:
        """Retrieves the current user temperature.

        Returns: user temperature in [Kelvin], or NaN if the user temperature is not available.
        """

        command = "GUT"
        response = self._execute_command(command)

        if response == "-0.100":
            # The instrument will return -0.100 if the user temperature is not available.
            user_stability = float("nan")
        else:
            user_stability = float(response)

        return user_stability

    @rpc_method
    def start_standby(self) -> None:
        """Start standby."""
        _logger.info("[%s] Starting standby.")

        command = "SSB"
        self._execute_command(command, check_ok=True)

    @rpc_method
    def start_cooldown(self) -> None:
        """Start the cooldown."""
        _logger.info("[%s] Starting the cooldown.")

        command = "SCD"
        self._execute_command(command, check_ok=True)

    @rpc_method
    def start_warmup(self) -> None:
        """Start the warmup."""
        _logger.info("[%s] Starting the warmup.")

        command = "SWU"
        self._execute_command(command, check_ok=True)

    @rpc_method
    def stop_system(self) -> None:
        """Stop either the standby, cooldown or warmup of the system."""
        _logger.info("[%s] Stopping the system.")

        command = "STP"
        self._execute_command(command, check_ok=True)

    def _internal_set_platform_heater_power(self, power: float) -> None:
        """Set platform heater to specified power."""

        # Power value has already been checked before calling this function.
        _logger.info("[%s] Setting platform heater to %f W", self._name, power)

        # Set "_platform_heater_on" before giving the command, so the flag
        # will be set properly even if the command gives an ambiguous response.
        if power > 0.0:
            self._platform_heater_on = True

        self._execute_command(f"SPHP {power:.4f}", check_ok=True)

        # Clear the "_platform_heater_on" flag after successfully turning off the heater.
        self._platform_heater_on = (power > 0.0)

    def _turn_off_platform_heater(self) -> None:
        """Turn off the platform heater.

        This function can be called:
         - by the timer thread, at the end of the specified heater burst duration;
         - after a failed attempt to set the platform heater;
         - when the instrument is closed.

        When invoked by the timer thread, this function runs inside that background thread.
        """

        # This function can run either inside the background timer thread or in
        # the main RPC thread. However it will never be invoked concurrently in both
        # threads because we stop the background timer before calling this function
        # when the instrument is closing.

        if self._platform_heater_on:

            # Turn off the platform heater.
            # Retry if necessary, because we really want to turn this heater off.
            for retry in range(self._MAX_TURN_OFF_HEATER_RETRY):
                _logger.info("[%s] Turning off platform heater", self._name)
                try:
                    self._internal_set_platform_heater_power(0.0)
                    break
                except (QMI_InstrumentException, QMI_TimeoutException) as exc:
                    _logger.error("[%s] Failed to turn off platform heater: %s",
                                  self._name, str(exc))
                    if retry >= self._MAX_TURN_OFF_HEATER_RETRY - 1:
                        # We are doomed. Report the error and find cover.
                        raise
                    # Short pause, then retry.
                    time.sleep(2.0)

    def _safe_set_platform_heater_power(self, power: float) -> None:
        """Set platform heater to specified power, or turn off the heater if an error occurs."""
        try:
            self._internal_set_platform_heater_power(power)
        except (QMI_InstrumentException, QMI_TimeoutException) as exc:
            _logger.error("[%s] Failed to set platform heater power: %s", self._name, str(exc))

            # Failed to change the heater power. This means the heater may still be at a dangerous level.
            # Turn the heater off to ensure a safe situation.
            self._turn_off_platform_heater()

            # Report the original exception.
            raise

    def _start_platform_heater_burst_timer(self, duration: float) -> None:
        """Start a background timer to turn off the platform heater."""
        assert self._platform_heater_timer_thread is None
        self._platform_heater_timer_thread = threading.Timer(duration, self._turn_off_platform_heater)
        _logger.debug("[%s] Starting burst timer for %f seconds", self._name, duration)
        self._platform_heater_timer_thread.start()

    def _cancel_platform_heater_burst_timer(self) -> None:
        """Cancel any currently running burst timer."""
        if self._platform_heater_timer_thread is not None:
            _logger.debug("[%s] Canceling burst timer", self._name)
            self._platform_heater_timer_thread.cancel()
            self._platform_heater_timer_thread.join()
            self._platform_heater_timer_thread = None

    @rpc_method
    def set_platform_heater_power(self, power: float) -> None:
        """Set the platform heater power.

        This function sets the platform heater to the specified power.
        The heater will remain at the requested power until explicitly turned off
        via another call to this function.

        This function only allows limited heater power levels. If you need more power,
        use `set_platform_heater_power_burst()` to set a high power level for a limited duration.

        Calling this function while a time-limited power burst is active, stops the burst timer and
        sets the new requested power without any time limit. The platform heater can always be turned off
        by calling this function and specifying 0 Watt power.

        Parameters:
            power:      Heater power in Watt (max 1 W).

        Raises:
            QMI_InstrumentException: If the Montana does not accept the command.
        """
        if not (0.0 <= power <= self._MAX_PLATFORM_HEATER_POWER):
            _logger.error("Trying to set %f W, but maximum is %d W", power, self._MAX_PLATFORM_HEATER_POWER)
            raise QMI_InvalidOperationException("Invalid platform heater power")

        # If a heater burst timer is running, cancel it.
        # This must be done before setting the heater, to make sure the timer will not interfere.
        # The timer is not needed anymore since we will set the heater to a safe level.
        self._cancel_platform_heater_burst_timer()

        # Set platform heater power, or turn off heater if an error occurs.
        self._safe_set_platform_heater_power(power)

    @rpc_method
    def set_platform_heater_power_burst(self, power: float, duration: float) -> None:
        """Set the platform heater to high power for a limited duration.

        This function sets the platform heater to the specified power, starts a timer,
        then returns while the timer keeps running in the background. If the timer reaches
        the specified duration, the heater power is automatically reset to 0 Watt.

        Calling this function while a background timer is running, restarts the timer for
        the specified new duration. It is thus possible to keep the heater at high power
        for a long time by repeatedly calling this function. A script that uses this technique,
        should also monitor temperatures and stop heating when temperatures reach a limit.

        A heating burst can always be stopped by calling this function (or `set_platform_heater_power`)
        and specifying 0 Watt power.

        WARNING: Setting the heater to high power for a long time can cause damage to
        the cryostat or the sample. Do not call this function unless you understand the risks.

        Parameters:
            power:      Heater power in Watt (max 10 W).
            duration:   Maximum duration for the specified power in seconds (max 30 s).

        Raises:
            QMI_InstrumentException: If the Montana does not accept the command.
        """

        if not (0.0 <= power <= self._MAX_PLATFORM_HEATER_BURST_POWER):
            _logger.error("Trying to set platform heater power burst of %f W for %f s, but maximum is %f W",
                          power, duration, self._MAX_PLATFORM_HEATER_BURST_POWER)
            raise QMI_InvalidOperationException("Invalid platform heater burst power")

        if (power > 0.0) and (duration < 0.0 or duration > self._MAX_PLATFORM_HEATER_BURST_DURATION):
            _logger.error("Trying to set platform heater power burst of %f W for %f s, but maximum is %f s",
                          power, duration, self._MAX_PLATFORM_HEATER_BURST_DURATION)
            raise QMI_InvalidOperationException("Invalid platform heater burst duration")

        # If a heater burst timer is running, cancel it.
        # This must be done before setting the heater, to make sure the timer will not interfere.
        self._cancel_platform_heater_burst_timer()

        # Set platform heater power, or turn off heater if an error occurs.
        self._safe_set_platform_heater_power(power)

        # If the heater is on, start a timer thread to turn it off after the specified duration.
        if power > 0:
            self._start_platform_heater_burst_timer(duration)



"""Toplevel package of the QMI framework.
"""

import sys
import logging
import os
import atexit


__version__ = "0.47.0"


# Check Python version.
if sys.version_info.major != 3:
    raise RuntimeError("QMI depends on Python 3. Your version: {}".format(sys.version_info))

# Set up basic logging functionality.
# This is done only when requested via the QMI_DEBUG environment variable.
# Normal logging setup is done during qmi.start().
if os.getenv("QMI_DEBUG"):
    import qmi.core.logging_init
    qmi.core.logging_init.start_logging(loglevel=logging.DEBUG,
                                        console_loglevel=logging.DEBUG)

# Direct call through .getLogger() to avoid making a package global variable.
logging.getLogger(__name__).debug("MODULE INIT: %s (enter)", __name__)

# Import symbols from specific QMI modules.
# This makes it more convenient to access these symbols from applications
# since they appear directly inside the qmi module.
from qmi.core.context_singleton import (
        context, start, stop, info, make_rpc_object, make_task, make_instrument, list_rpc_objects,
        show_rpc_objects, show_instruments, show_tasks, show_contexts, show_network_contexts,
        get_rpc_object, get_instrument, get_task, get_configured_contexts
    )

from qmi.core.object_registry import ObjectRegistry as _ObjectRegistry

object_registry = _ObjectRegistry()

atexit.register(lambda: object_registry.report(os.getenv("QMI_DEBUG") is not None))

logging.getLogger(__name__).debug("MODULE INIT: %s (leave)", __name__)



"""Instrument driver for the JPE CPSC piezo stage controller."""

import logging
import os
import sys
from subprocess import PIPE, CompletedProcess, Popen, run
from typing import Any, List, Optional, Iterable, NamedTuple

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException, QMI_UsageException
from qmi.core.instrument import QMI_Instrument, QMI_InstrumentIdentification
from qmi.core.rpc import rpc_method

# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


class StatusPositionControl(NamedTuple):
    """A named tuple with status and position error information."""
    ENABLED: int
    FINISHED: int
    INVALID_SP1: int
    INVALID_SP2: int
    INVALID_SP3: int
    POS_ERROR1: int
    POS_ERROR2: int
    POS_ERROR3: int


class JPE_CPSC(QMI_Instrument):
    """Instrument driver for the JPE CPSC piezo stage controller.

    Note that the piezo stage controller will only hold its actuator positions
    while the connection to the instrument is open. Closing the connection
    will automatically return the controller to its free state.

    This driver communicates with the piezo stage controller via
    a helper program "cacli.exe".

    The helper program is a command-line Windows application.
    On Linux, it can run under Wine.
    """

    # Location of cpsc program.
    DEFAULT_CPSC_DIR = "C:\\HardwareControl\\CPSC_v7.3.20201222"

    CPSC_CMD = "cacli.exe"

    # USD vendor ID : product ID of the piezo stage controller.
    USB_ID = "045E:FFFF"

    def __init__(self, context: QMI_Context, name: str, serial_number: str, cpsp_dir: str = DEFAULT_CPSC_DIR) -> None:
        """Initialize the driver.

        Arguments:
            name: Name for this instrument instance.
            serial_number: Serial number of the controller.
            cpsp_dir: Path to directory where cacli.exe is installed.

        Raises:
            QMI_InstrumentException: If the CPSC program directory {cpsp_dir} is not found
        """
        if not os.path.isdir(cpsp_dir):
            raise QMI_InstrumentException(f"CPSC program directory {cpsp_dir} not found")
        super().__init__(context, name)
        if serial_number.startswith("@"):
            self._serial_number = serial_number

        else:
            self._serial_number = "@" + serial_number

        self._cpsp_dir = cpsp_dir
        self._cpsp_cmd = os.path.join(cpsp_dir, self.CPSC_CMD)
        self._proc_popen = None  # type: Optional[Popen]
        self._proc_run = None  # type: Optional[CompletedProcess]
        self._version_string = ""

    def _start_helper_wine(self) -> None:
        """Start the WinE. WARNING: NOT TESTED WITH HARDWARE!"""
        assert self._proc_popen is None
        env: dict = {}
        env.update(os.environ)
        env["WINEDLLPATH"] = "."
        env["FTDID"] = self.USB_ID
        self._proc_popen = Popen(args="wine", stdin=PIPE, stdout=PIPE, cwd=self._cpsp_dir, env=env)
        assert self._proc_popen.stdin is not None
        assert self._proc_popen.stdout is not None

    def _stop_helper(self) -> None:
        """Stop the helper program."""
        assert self._proc_popen is not None
        assert self._proc_popen.stdin is not None
        assert self._proc_popen.stdout is not None
        self._proc_popen.stdin.close()
        self._proc_popen.wait()
        self._proc_popen.stdout.close()
        self._proc_popen = None
        self._version_string = ""

    def _send_cmd(self, cmd: str, args: Optional[Iterable] = None) -> None:
        """Send command to helper program and check response.

        Arguments:
            cmd: The protocol command to send (without serial number)
            args: A list of input parameters. Can be in any format that can be cast as string.
        """
        if args:
            _logger.debug("Sending command {} with args {} to JPE CPSC {}".format(
                self._name, ",".join(map(str, args)), self._serial_number)
            )

        else:
            _logger.debug("Sending command {} without arguments to JPE CPSC {}".format(
                self._name, self._serial_number)
            )

        if sys.platform.startswith("linux"):
            assert self._proc_popen is not None
            assert self._proc_popen.stdin is not None
            # Add serial number at the start
            cmd = self._cpsp_cmd + " " + self._serial_number + " " + cmd
            if args:
                for arg in args:
                    cmd += " " + str(arg)

            self._proc_popen.stdin.write(cmd.encode("ascii") + b"\r\n")
            self._proc_popen.stdin.flush()

        elif sys.platform == "darwin":
            raise QMI_UsageException("Mac OS X not supported.")

        else:
            # We need to make argument list. Start with exe, then serial number followed by command and args
            arg_list = [self._cpsp_cmd, self._serial_number, cmd]
            if args:
                for arg in args:
                    arg_list.append(str(arg))

            self._proc_run = run(arg_list, stdin=PIPE, stdout=PIPE, encoding="utf-8", check=False)

    def _recv_response(self) -> str:
        """Read response from helper program.

        Raises:
            QMI_InstrumentException: If one of the read stdout lines is unexpectedly empty or None.
        """
        s: str = ""
        if sys.platform.startswith("linux"):
            assert self._proc_popen is not None
            assert self._proc_popen.stdout is not None
            lines = self._proc_popen.stdout.readlines()
            _logger.debug("recv: {}".format(lines))
            for line in lines:
                if not line:
                    raise QMI_InstrumentException("Unexpected end of input from helper program")

                line = line.decode()
                s += line

        else:
            s = self._proc_run.stdout  # type: ignore

        _logger.debug("Received response {} from JPE CPSC {}".format(s, self._serial_number))
        return s

    def _check_response(self, expected: str) -> None:
        """Check the obtained response against expected response.

        Arguments:
             expected: The expected response string.

        Raises:
            QMI_InstrumentException: If response start with "ERROR" message or response does not hold
            the expected string.
        """
        s = self._recv_response()
        if expected in s:
            return

        elif s.startswith("ERROR"):
            raise QMI_InstrumentException(f"Error from {self._name}: {s}")

        else:
            raise QMI_InstrumentException(f"Unexpected response from {self._name}: {s}. Expected {expected}")

    def _get_cli_version(self) -> str:
        """Get CLI version number.

        Returns:
            Response to the query as string.
        """
        self._send_cmd("/VER")
        return self._recv_response()

    @rpc_method
    def open(self) -> None:
        """The 'open()' function override. Starts the correct os-specific helper program."""
        self._check_is_closed()

        if sys.platform.startswith("linux"):  # TODO: Test with hardware on Linux if this really works.
            _logger.info("[%s] Starting helper program", self._name)
            self._start_helper_wine()

        super().open()

    @rpc_method
    def close(self) -> None:
        """The 'close()' function override. Stops the helper program and closes the instrument."""
        self._check_is_open()
        if sys.platform.startswith("linux"):
            _logger.info("[%s] Stopping helper program", self._name)
            self._stop_helper()

        _logger.info("[%s] the instrument", self._name)
        super().close()

    @rpc_method
    def get_idn(self) -> QMI_InstrumentIdentification:
        """Read instrument type and version.

        Returns:
            QMI_InstrumentIdentification instance.
        """
        self._check_is_open()
        self._version_string = self._get_cli_version()
        return QMI_InstrumentIdentification(
            vendor="JPE", model="CPSC", serial=self._serial_number, version=self._version_string
        )

    @rpc_method
    def get_info_all_modules(self) -> List[str]:
        """Get information about installed modules.

        Returns:
            modlist: Listing of module info
        """
        self._check_is_open()
        self._send_cmd("MODLIST")
        modlist = self._recv_response().split(",")
        return modlist

    @rpc_method
    def get_info_module(self, address: int) -> str:
        """Get information about given module.

        Arguments:
            address: 1 to 6. Address of module corresponding to controller slot.
        Returns:
            info string about given module
        """
        self._check_is_open()
        self._send_cmd("Desc", [address])
        return self._recv_response().strip()

    @rpc_method
    def request_fail_safe_state(self, address: int) -> str:
        """Reguest fail-safe state from a specific module. In practice - check for errors.

        Arguments:
            address: 1 to 6. Address of module corresponding to controller slot.

        Returns:
            the response error code string
        """
        self._check_is_open()
        self._send_cmd("GFS", [address])
        return self._recv_response().strip()

    @rpc_method
    def move(
            self,
            address: int,
            direction: int,
            freq: int = 10,
            rss: int = 100,
            steps: int = 0,
            temp: int = 295,
            stage: str = "CBS10-RLS",
            df: float = 1.0,
    ) -> None:
        """Move an actuator with specific parameters

        Arguments:
            address: 1 to 6. Address of module corresponding to controller slot.
            direction: 0 to 1. Direction of movement: set to 1 for positive movement and 0 (zero) for
                    negative movement
            freq: 1 to 600 Step frequency input. Value is in Hertz [Hz] (numerical values only).
            rss: 1 to 100 (Relative) actuator step size parameter input. Value is a percentage [%]
                    (numerical values only)
            steps: 0 to 50000 Number of actuation steps. Range 0 to 50000, where 0 is used for
                        infinite move (use STP command to stop actuator movement).
            temp: 0 to 300 Set this parameter to the temperature of the environment in which the
                    actuator is used. Input is in Kelvin [K] (numerical values only).
            stage: Sets specific internal drive parameters for the type of actuator or system
                        attached to that particular channel of that particular module set by [ADDR] and [CH].
            df: Optional[float] - 0.1 to 3.0 Drive factor (numerical values only). In normal operating conditions,
                    set this value to 1 (or 1.0).
        """
        self._check_is_open()
        args: List[Any] = [address, direction, freq, rss, steps, temp, stage, df]

        self._send_cmd("MOV", args)
        self._check_response("Actuating the stage.")

    @rpc_method
    def stop(self, address: int) -> None:
        """Stops movement of an actuator (MOV command), disables external input mode (EXT command) or disables
        scan mode (SDC command).

        Arguments:
            address: 1 to 6. Address of module corresponding to controller slot.
        """
        self._check_is_open()
        self._send_cmd("STP", [address])
        self._check_response("Stopping the stage.")

    @rpc_method
    def enable_scan_mode(self, address: int, value: int) -> None:
        """Enables and sets the scan mode for CADM2.

        Arguments:
            address: 1 to 6. Address of module corresponding to controller slot.
            value: A numerical value in between 0 and 1023 (10-bit value) where zero represents ~0[V] output
         (-30[V] in respect to REF) and the maximum value represents ~150[V] output (+120[V] in respect to REF).
        """
        self._check_is_open()
        self._send_cmd("SDC", [address, value])
        self._check_response("Scan mode enabled.")

    @rpc_method
    def use_external_input(
            self,
            address: int,
            direction: int,
            freq: int = 10,
            rss: int = 100,
            temp: int = 295,
            stage: str = "CBS10-RLS",
            df: float = 1.0,
    ) -> None:
        """To use the CADM2 in Flexdrive mode, it is required to set the module in external (analog) input mode
        prior to using Flexdrive.

        Arguments:
             address: 1 to 6. Address of module corresponding to controller slot.
             direction: 0 to 1. Direction of movement: set to 1 for positive movement and 0 (zero) for
                    negative movement
             freq: 1 to 600 Step frequency input. Value is in Hertz [Hz] (numerical values only).
             rss: 1 to 100 (Relative) actuator step size parameter input. Value is a percentage [%]
                    (numerical values only)
             temp: 0 to 300 Set this parameter to the temperature of the environment in which the
                    actuator is used. Input is in Kelvin [K] (numerical values only).
             stage:  Sets specific internal drive parameters for the type of actuator or system
                        attached to that particular channel of that particular module set by [ADDR] and [CH].
             df: 0.1 to 3.0 Drive factor (numerical values only). In normal operating conditions, set
                    this value to 1 (or 1.0).
        """
        self._check_is_open()
        args = []
        for arg in [address, direction, freq, rss, temp, stage, df]:
            if not isinstance(arg, type(None)):
                args.append(arg)

        self._send_cmd("EXT", args)
        self._check_response("External mode enabled.")

    @rpc_method
    def get_current_position(self, address: int, chan: int, stage: str = "CBS10-RLS") -> float:
        """Get current position of a Resistive Linear Sensor (RLS) connected to a specific channel [CH] of the RSM
        module.
        Arguments:
            address: 1 to 6. Address of module corresponding to controller slot.
            chan: 1 to 3. Specific channel number.
            stage: stage name, e.g. "CLA2601", default "CBS10-RLS".
        Raises:
            QMI_InstrumentException: If the response cannot be cast into a float.
        Returns:
            cur_pos: Current position in [m].
        """
        self._check_is_open()
        self._send_cmd("PGV", [address, chan, stage])
        response = self._recv_response().strip()
        try:
            cur_pos = float(response)

        except Exception as exc:
            raise QMI_InstrumentException(
                f"Erroneous response from position query SDC {address} {chan} {stage}: {response}"
            ) from exc

        return cur_pos

    @rpc_method
    def get_current_position_of_all_3_channels(self, address: int, stage: str = "CBS10-RLS") -> List[float]:
        """Get current position of all three channels of the RSM module.

        Arguments:
            address: 1 to 6. Address of module corresponding to controller slot.
            stage: stage name, e.g. "CLA2601", default "CBS10-RLS".
        Raises:
            QMI_InstrumentException: If (one of the) response entries cannot be cast into a float.
        Returns:
            cur_pos: current positions in [m].
        """
        self._check_is_open()
        cur_pos = []
        self._send_cmd("PGVA", [address, stage, stage, stage])
        response = self._recv_response().strip()
        try:
            for rspv in response.split(","):
                cur_pos.append(float(rspv))

        except Exception as exc:
            raise QMI_InstrumentException(
                f"Erroneous response from position query PGVA {address} {stage}: {response}"
            ) from exc

        return cur_pos

    @rpc_method
    def set_negative_end_stop(self, address: int, chan: int) -> None:
        """Set the current position of a Resistive Linear Sensor (RLS) connected to channel [CH] of the RSM to be
        the negative end-stop.

        Arguments:
            address: 1 to 6. Address of module corresponding to controller slot.
            chan: 1 to 3. Specific channel number.
        """
        self._check_is_open()
        self._send_cmd("MIS", [address, chan])
        self._check_response("Minimum position set.")

    @rpc_method
    def set_positive_end_stop(self, address: int, chan: int) -> None:
        """Set the current position of a Resistive Linear Sensor (RLS) connected to channel [CH] of the RSM to be
        the positive end-stop.

        Arguments:
            address: 1 to 6. Address of module corresponding to controller slot.
            chan: 1 to 3. Specific channel number.
        """
        self._check_is_open()
        self._send_cmd("MAS", [address, chan])
        self._check_response("Maximum position set.")

    @rpc_method
    def read_negative_end_stop(self, address: int, chan: int, stage: str = "CBS10-RLS") -> float:
        """Read the current value of the negative end-stop parameter set for a specific channel [CH] of an RSM.

        Arguments:
            address: 1 to 6. Address of module corresponding to controller slot.
            chan: 1 to 3. Specific channel number.
            stage: Stage name, e.g. "CLA2601", default "CBS10-RLS".
        Raises:
            QMI_InstrumentException: If the response cannot be cast into a float.
        Returns:
            cur_val: The current value of the parameter in [m].
        """
        self._check_is_open()
        self._send_cmd("MIR", [address, chan, stage])
        response = self._recv_response().strip()
        try:
            cur_val = float(response)

        except Exception as exc:
            raise QMI_InstrumentException(
                f"Erroneous response from end-stop query MIR {address} {chan} {stage}: {response}"
            ) from exc

        return cur_val

    @rpc_method
    def read_positive_end_stop(self, address: int, chan: int, stage: str = "CBS10-RLS") -> float:
        """Read the current value of the positive end-stop parameter set for a specific channel [CH] of an RSM.

        Arguments:
            address: 1 to 6. Address of module corresponding to controller slot.
            chan: 1 to 3. Specific channel number.
            stage: stage name, e.g. "CLA2601", default "CBS10-RLS".
        Raises:
            QMI_InstrumentException: If the response cannot be cast into a float.
        Returns:
            cur_val: the current value of the parameter in [m].
        """
        self._check_is_open()
        self._send_cmd("MAR", [address, chan, stage])
        response = self._recv_response().strip()
        try:
            cur_val = float(response)

        except Exception as exc:
            raise QMI_InstrumentException(
                f"Erroneous response from end-stop query MAR {address} {chan} {stage}: {response}"
            ) from exc

        return cur_val

    @rpc_method
    def reset_end_stops(self, address: int, chan: int):
        """Reset the current values of the negative and positive end-stop parameters set for a specific channel [CH]
        of an RSM.

        Arguments:
            address: 1 to 6. Address of module corresponding to controller slot.
            chan: 1 to 3. Specific channel number.
        """
        self._check_is_open()
        self._send_cmd("MMR", [address, chan])
        self._check_response("Minimum and maximum end-stops reset.")

    @rpc_method
    def set_excitation_duty_cycle(self, address: int, duty: int) -> None:
        """Set the duty cycle of the sensor excitation signal of the RSM.

        Arguments:
            address: 1 to 6. Address of module corresponding to controller slot.
            duty: Value is in [%] and can be set to 0 (zero) or from 10 to (default) 100.
        """
        self._check_is_open()
        self._send_cmd("EXS", [address, duty])
        self._check_response("Excitation duty cycle set.")

    @rpc_method
    def read_excitation_duty_cycle(self, address: int) -> int:
        """Read the duty cycle of the sensor excitation signal of the RSM.

        Arguments:
            address: 1 to 6. Address of module corresponding to controller slot.
        Raises:
            QMI_InstrumentException: If the response cannot be cast into an int.
        Returns:
            cur_val: Response value is in [%].
        """
        self._check_is_open()
        self._send_cmd("EXR", [address])
        response = self._recv_response().strip()
        try:
            cur_val = int(response)

        except Exception as exc:
            raise QMI_InstrumentException(
                f"Erroneous response from excitation duty cycle query EXR {address}: {response}"
            ) from exc

        return cur_val

    @rpc_method
    def save_rsm_settings(self, address: int) -> None:
        """Store the current values of the following parameters of the RSM to the non-volatile memory of the
        controller: excitation duty cycle (EXS), negative end stop (MIS) and positive end-stop (MAS).

        Arguments:
            address: 1 to 6. Address of module corresponding to controller slot.
        """
        self._check_is_open()
        self._send_cmd("RSS", [address])
        self._check_response("Settings stored in flash.")

    @rpc_method
    def get_current_counter_value(self, address: int, chan: int) -> int:
        """Request the counter valuea of a Cryo Optical Encoder (COE) connected to a specific channel [CH] of the
        OEM2 module.

        Arguments:
            address: 1 to 6. Address of module corresponding to controller slot.
            chan: 1 to 3. Specific channel number.
        Raises:
            QMI_InstrumentException: If the response cannot be cast into an int.
        Returns:
            cur_val: Response value is in [counter ticks].
        """
        self._check_is_open()
        self._send_cmd("CGV", [address, chan])
        response = self._recv_response().strip()
        try:
            cur_val = int(response)

        except Exception as exc:
            raise QMI_InstrumentException(
                f"Erroneous response from current counter value query CGV {address} {chan}: {response}"
            ) from exc

        return cur_val

    @rpc_method
    def reset_counter_value(self, address: int, chan: int) -> None:
        """Resets the counter (to zero) for a specific cryo optical encoder connected to a specific channel [CH] of
        the OEM2 module.

        Arguments:
            address: 1 to 6. Address of module corresponding to controller slot.
            chan: 1 to 3. Specific channel number.
        """
        self._check_is_open()
        self._send_cmd("CSZ", [address, chan])
        self._check_response("Position counter set to 0.")

    @rpc_method
    def get_current_encoder_signal_value(self, address: int, chan: int) -> int:
        """Request the (raw) encoder signal value of a Cryo Optical Encoder (COE) connected to a specific channel
        [CH] of the OEM2 module.

        Arguments:
            address: 1 to 6. Address of module corresponding to controller slot.
            chan: 1 to 3. Specific channel number.
        Raises:
            QMI_InstrumentException: If the response cannot be cast into an int.
        Returns:
            cur_val: Return value is a (unitless) number between [0] and [255].
        """
        self._check_is_open()
        self._send_cmd("DGV", [address, chan])
        response = self._recv_response().strip()
        try:
            cur_val = int(response)

        except Exception as exc:
            raise QMI_InstrumentException(
                f"Erroneous response from current encoder signal value query DGV {address} {chan}: {response}"
            ) from exc

        return cur_val

    @rpc_method
    def auto_oem_calibration(self, address: int, chan: int, cadm2_address: int, temp: int, stage: str = "CBS10-RLS")\
            -> None:
        """Command to initiate an automatic calibration procedure for a specific encoder connected to channel
        [CH] of an OEM2.

        Arguments:
            address: 1 to 6. Address of module corresponding to controller slot.
            chan: 1 to 3. Specific channel number.
            cadm2_address: 1 to 6. Address of module corresponding to CADM2 slot.
            temp: 0 to 300 Set this parameter to the temperature of the environment in which the
                    actuator is used. Input is in Kelvin [K] (numerical values only).
            stage: stage name, e.g. "CLA2601", default "CBS10-RLS".
        """
        self._check_is_open()
        self._send_cmd("OEMC", [address, chan, cadm2_address, temp, stage])
        self._check_response("Channel calibrated.")

    @rpc_method
    def request_calibration_values(self, address: int, chan: int) -> List[int]:
        """Request the Detector Gain setting [GAIN], Upper Threshold value [UT] and Lower Threshold value [LT]
        set to a specific channel [CH] of the OEM2 module.

        Arguments:
            address: 1 to 6. Address of module corresponding to controller slot.
            chan: 1 to 3. Specific channel number.
        Raises:
            QMI_InstrumentException: If (one of the) response entries cannot be cast into an int.
        Returns:
            cal_vals: Three values in between [1] and [255].
        """
        self._check_is_open()
        cal_vals = []
        self._send_cmd("MLS", [address, chan])
        response = self._recv_response().strip()
        try:
            for rspv in response.split(","):
                cal_vals.append(int(rspv))

        except Exception as exc:
            raise QMI_InstrumentException(
                f"Erroneous response from calibration value query MLS {address} {chan}: {response}"
            ) from exc

        return cal_vals

    @rpc_method
    def set_detector_gain(self, address: int, chan: int, gain: int) -> None:
        """Set a Detector Gain [GAIN] for a specific channel [CH] of the OEM2 module.

        Arguments:
            address: 1 to 6. Address of module corresponding to controller slot.
            chan: 1 to 3. Specific channel number.
            gain: a value in between [1] and [255].
        """
        self._check_is_open()
        self._send_cmd("DSG", [address, chan, gain])
        self._check_response("Detector gain set.")

    @rpc_method
    def set_upper_threshold(self, address: int, chan: int, ut: int) -> None:
        """Set a Detector Upper Threshold value [UT] for a specific channel [CH] of the OEM2 module.

        Arguments:
            address: 1 to 6. Address of module corresponding to controller slot.
            chan: 1 to 3. Specific channel number.
            ut: a value in between [1] and [255]. Must be higher than Lower Threshold [LT] value.
        """
        self._check_is_open()
        self._send_cmd("DSH", [address, chan, ut])
        self._check_response("Detector upper threshold set.")

    @rpc_method
    def set_lower_threshold(self, address: int, chan: int, lt: int) -> None:
        """Set a Detector Lower Threshold value [LT] for a specific channel [CH] of the OEM2 module.

        Arguments:
            address: 1 to 6. Address of module corresponding to controller slot.
            chan: 1 to 3. Specific channel number.
            lt: a value in between [1] and [255]. Must be lower than Upper Threshold [UT] value.
        """
        self._check_is_open()
        self._send_cmd("DSL", [address, chan, lt])
        self._check_response("Detector lower threshold set.")

    @rpc_method
    def save_detector_settings(self, address) -> None:
        """Store Detector Gain [GAIN], Detector Upper Threshold value [UT] and Detector Lower Threshold value [LT]
         values in nonvolatile memory.

        Arguments:
            address: 1 to 6. Address of module corresponding to controller slot.
        """
        self._check_is_open()
        self._send_cmd("MSS", [address])
        self._check_response("Detector settings stored in flash.")

    @rpc_method
    def enable_servodrive(
            self, stage1: str, freq1: int, stage2: str, freq2: int, stage3: str, freq3: int, df: int, temp: int
    ) -> None:
        """Enable servo drive control loop.

        Arguments:
            stage1: Sets specific internal drive parameters for the type of actuator or system
                        attached to that particular channel of that particular module set by [ADDR] and [CH].
            freq1: 1 to 600 Step frequency input. Value is in Hertz [Hz] (numerical values only).
            stage2: Sets specific internal drive parameters for the type of actuator or system
                        attached to that particular channel of that particular module set by [ADDR] and [CH].
            freq2: 1 to 600 Step frequency input. Value is in Hertz [Hz] (numerical values only).
            stage3: Sets specific internal drive parameters for the type of actuator or system
                        attached to that particular channel of that particular module set by [ADDR] and [CH].
            freq3: 1 to 600 Step frequency input. Value is in Hertz [Hz] (numerical values only).
            df: 0.1 to 3.0 Drive factor (numerical values only). In normal operating conditions, set
                    this value to 1 (or 1.0).
            temp: 0 to 300 Set this parameter to the temperature of the environment in which the
                        actuator is used. Input is in Kelvin [K] (numerical values only).
        """
        self._check_is_open()
        self._send_cmd("FBEN", [stage1, freq1, stage2, freq2, stage3, freq3, df, temp])
        self._check_response("Control loop enabled.")

    @rpc_method
    def disable_servodrive(self) -> None:
        """Disable servo drive control loop."""
        self._check_is_open()
        self._send_cmd("FBXT")
        self._check_response("Control loop disabled.")

    @rpc_method
    def move_to_setpoint(self, sp1: float, sp2: float, sp3: float, enable_abs_pos: int) -> None:
        """Move actuators to a set point position.

        Arguments:
            sp1: setpoint value in [m] for linear, in [rad] for rotational type actuators.
            sp2: setpoint value in [m] for linear, in [rad] for rotational type actuators.
            sp3: setpoint value in [m] for linear, in [rad] for rotational type actuators.
            enable_abs_pos: 1 to enable absolute positioning, otherwise set to 0 (zero).
        """
        self._check_is_open()
        self._send_cmd("FBCS", [sp1, enable_abs_pos, sp2, enable_abs_pos, sp3, enable_abs_pos])
        self._check_response("Control loop setpoints set.")

    @rpc_method
    def emergency_stop(self) -> None:
        """Immediate stop of all actuators. The control loop will be aborted and the actuators will stop at
        their current location."""
        self._check_is_open()
        self._send_cmd("FBES")
        self._check_response("Control loop emergency stop enabled.")

    @rpc_method
    def get_status_position_control(self) -> StatusPositionControl:
        """This command receives a (comma-separated) list with status and position error information. The list is then
        turned into a named tuple which is returned.

        Returns:
            [ENABLED] [FINISHED] [INVALID SP1] [INVALID SP2] [INVALID SP3] [POS ERROR1] [POS ERROR2]
            [POS ERROR3]
        """
        self._check_is_open()
        self._send_cmd("FBST")
        status = self._recv_response().strip()
        return StatusPositionControl(*map(int, status.split(",")))



"""Test for the USB1808x driver."""
import struct

from math import isnan

from unittest import TestCase
from unittest.mock import MagicMock
from unittest.mock import patch
from unittest.mock import call

from typing import cast

from dataclasses import dataclass


from qmi.core.transport import QMI_Transport
from qmi.core.scpi_protocol import ScpiProtocol
from qmi.instruments.mcc.usb1808x import MCC_USB1808X
from qmi.core.exceptions import QMI_InstrumentException


@dataclass
class TestMeta:
    """Test meta data."""

    name: MagicMock
    id: MagicMock
    uldaq: MagicMock
    super: MagicMock
    import_module: MagicMock
    instr: MCC_USB1808X


@dataclass
class TestDeviceMeta:
    """Test device data."""

    descriptor: MagicMock
    device: MagicMock
    dio_device: MagicMock
    ai_device: MagicMock
    ao_device: MagicMock


@dataclass
class TestUldaqDeviceDescriptor:
    """Test device descriptor data."""

    product_name: str
    unique_id: MagicMock()


class TestMCC_USB1808X(TestCase):
    """Testcase for the MCC_USB1808X class.

    Decided not to check assertions for self._device and self._*_device in
    every function. These assertions are unnesecary in the code of the
    USB1808x driver, self._check_is_open() already covers this check.
    """

    def setUp(self):
        mock_name = MagicMock()
        mock_id = MagicMock()
        mock_uldaq = MagicMock()
        mock_super = MagicMock()
        mock_import_module = MagicMock()

        self._patcher_import_module = patch(
            "qmi.instruments.mcc.usb1808x._import_modules",
            mock_import_module,
        )
        self._patcher_import_module.start()

        self._patcher_uldaq = patch("qmi.instruments.mcc.usb1808x.uldaq", mock_uldaq)
        self._patcher_uldaq.start()

        instr = MCC_USB1808X(MagicMock(), mock_name, mock_id)

        self._patcher_super = patch("qmi.instruments.mcc.usb1808x.super", mock_super)
        self._patcher_super.start()

        self._meta = TestMeta(
            instr=cast(MCC_USB1808X, instr),
            name=mock_name,
            id=mock_id,
            uldaq=mock_uldaq,
            super=mock_super,
            import_module=mock_import_module,
        )

    def tearDown(self):
        self._meta = None
        self._patcher_import_module.stop()
        self._patcher_uldaq.stop()
        self._patcher_super.stop()

    def test_init(self):
        """MCC_USB1808X.__init__(), happy flow."""
        self._meta.import_module.assert_called_once_with()

    def test_list_instruments(self):
        """MCC_USB1808X.list_instruments(), happy flow."""
        mock_unique_id_1 = MagicMock()
        mock_unique_id_2 = MagicMock()
        mock_unique_id_3 = MagicMock()
        mock_resp = [
            TestUldaqDeviceDescriptor("USB-1808X", mock_unique_id_1),
            TestUldaqDeviceDescriptor("NOT", mock_unique_id_2),
            TestUldaqDeviceDescriptor("USB-1808X", mock_unique_id_3),
        ]

        self._meta.uldaq.get_daq_device_inventory = MagicMock(return_value=mock_resp)
        rt_val = self._meta.instr.list_instruments()
        self._meta.import_module.assert_has_calls([call(), call()])
        self._meta.uldaq.get_daq_device_inventory.assert_called_once_with(
            self._meta.uldaq.InterfaceType.USB
        )
        self.assertEqual(rt_val[0], mock_unique_id_1)
        self.assertEqual(rt_val[1], mock_unique_id_3)

    def _mock_device(self):
        """Initializes self._meta.uldaq with a mocked device."""
        mock_descriptor = TestUldaqDeviceDescriptor("USB-1808X", self._meta.id)
        self._meta.uldaq.get_daq_device_inventory = MagicMock(
            return_value=[mock_descriptor]
        )

        mock_device = MagicMock()
        self._meta.uldaq.DaqDevice = MagicMock(return_value=mock_device)

        mock_dio_device = MagicMock()
        mock_ai_device = MagicMock()
        mock_ao_device = MagicMock()
        mock_device.get_dio_device = MagicMock(return_value=mock_dio_device)
        mock_device.get_ai_device = MagicMock(return_value=mock_ai_device)
        mock_device.get_ao_device = MagicMock(return_value=mock_ao_device)

        return TestDeviceMeta(
            descriptor=mock_descriptor,
            device=mock_device,
            dio_device=mock_dio_device,
            ai_device=mock_ai_device,
            ao_device=mock_ao_device,
        )

    def test_open(self):
        """MCC_USB1808X.open(), happy flow."""
        self._meta.instr._check_is_closed = MagicMock()

        device_meta = self._mock_device()

        self._meta.instr.open()

        self._meta.instr._check_is_closed.assert_called_once_with()
        device_meta.device.connect.assert_called_once_with()
        self._meta.uldaq.DaqDevice.assert_called_once_with(device_meta.descriptor)
        self.assertEqual(self._meta.instr._device, device_meta.device)
        self.assertEqual(self._meta.instr._dio_device, device_meta.dio_device)
        self.assertEqual(self._meta.instr._ai_device, device_meta.ai_device)
        self.assertEqual(self._meta.instr._ao_device, device_meta.ao_device)
        self._meta.super().open.assert_called_once_with()

    def test_open_no_device(self):
        """MCC_USB1808X.open(), no device found exception."""
        self._meta.instr._check_is_closed = MagicMock()

        device_meta = self._mock_device()
        device_meta.descriptor.product_name = "NOT"
        device_meta.descriptor.id = "NOT"

        with self.assertRaises(QMI_InstrumentException):
            self._meta.instr.open()

    def test_open_connect_error(self):
        """MCC_USB1808X.open(), device.connect() exception."""
        self._meta.instr._check_is_closed = MagicMock()

        device_meta = self._mock_device()
        device_meta.device.connect = MagicMock(side_effect=Exception)

        with self.assertRaises(Exception):
            self._meta.instr.open()

        device_meta.device.release.assert_called_once_with()

    def test_open_device_get_dio_device_error(self):
        """MCC_USB1808X.open(), device.get_dio_device() exception."""
        self._meta.instr._check_is_closed = MagicMock()

        device_meta = self._mock_device()
        device_meta.device.get_dio_device = MagicMock(side_effect=Exception)

        with self.assertRaises(Exception):
            self._meta.instr.open()

        device_meta.device.disconnect.assert_called_once_with()
        device_meta.device.release.assert_called_once_with()

    def test_close(self):
        """MCC_USB1808X.close(), happy flow."""
        self._meta.instr._check_is_open = MagicMock()
        device_meta = self._mock_device()
        self._meta.instr._device = (
            device_meta.device
        )  # want to make sure disconnect and release arent called through open() thus set the device directly.
        self._meta.instr.close()
        device_meta.device.disconnect.assert_called_once_with()
        device_meta.device.release.assert_called_once_with()
        self._meta.instr._check_is_open.assert_called_once_with()
        self._meta.super().close.assert_called_once_with()

    def test_get_idn(self):
        """MCC_USB1808X.get_idn(), happy flow."""
        self._meta.instr._check_is_open = MagicMock()
        device_meta = self._mock_device()
        self._meta.instr.open()
        rt_val = self._meta.instr.get_idn()
        self._meta.instr._check_is_open.assert_called_once_with()
        device_meta.device.get_config().get_version.assert_called_once_with(
            self._meta.uldaq.DevVersionType.FW_MAIN
        )
        self.assertEqual(rt_val.vendor, "Measurement Computing")
        self.assertEqual(rt_val.model, device_meta.device.get_descriptor().product_name)
        self.assertEqual(rt_val.serial, device_meta.device.get_descriptor().unique_id)
        self.assertEqual(rt_val.version, device_meta.device.get_config().get_version())

    def test_dio_num_channels(self):
        """MCC_USB1808X.get_dio_num_channels(), happy flow."""
        self._meta.instr._check_is_open = MagicMock()
        device_meta = self._mock_device()
        self._meta.instr.open()
        rt_val = self._meta.instr.get_dio_num_channels()
        device_meta.dio_device.get_info.assert_called_once_with()
        device_meta.dio_device.get_info().get_port_info.assert_called_once_with(
            self._meta.uldaq.DigitalPortType.AUXPORT
        )
        self.assertEqual(
            rt_val, device_meta.dio_device.get_info().get_port_info().number_of_bits
        )
        self._meta.instr._check_is_open.assert_called_once_with()

    def test_get_dio_direction(self):
        """MCC_USB1808X.get_dio_direction(), happy flow."""
        self._meta.instr._check_is_open = MagicMock()
        device_meta = self._mock_device()
        self._meta.instr.open()
        mock_directions = [
            self._meta.uldaq.DigitalDirection.OUTPUT,
            self._meta.uldaq.DigitalDirection.INPUT,
            self._meta.uldaq.DigitalDirection.OUTPUT,
        ]
        device_meta.dio_device.get_config().get_port_direction = MagicMock(
            return_value=mock_directions
        )
        device_meta.dio_device.get_config.reset_mock()
        rt_val = self._meta.instr.get_dio_direction()
        device_meta.dio_device.get_config.assert_called_once_with()
        device_meta.dio_device.get_config().get_port_direction.assert_called_once_with(
            self._meta.uldaq.DigitalPortType.AUXPORT
        )
        # OUTPUT is represented as a True, INPUT as a False
        self.assertEqual(rt_val[0], True)
        self.assertEqual(rt_val[1], False)
        self.assertEqual(rt_val[2], True)
        self._meta.instr._check_is_open.assert_called_once_with()

    def test_set_dio_direction(self):
        """MCC_USB1808X.set_dio_direction(), happy flow."""
        device_meta = self._mock_device()
        self._meta.instr.open()
        map = [
            {"output": True, "direction": self._meta.uldaq.DigitalDirection.OUTPUT},
            {"output": False, "direction": self._meta.uldaq.DigitalDirection.INPUT},
        ]
        for item in map:
            self._meta.instr._check_is_open = MagicMock()
            mock_channel = MagicMock()
            mock_output = item["output"]
            device_meta.dio_device.d_config_bit = MagicMock()
            self._meta.instr.set_dio_direction(mock_channel, mock_output)
            device_meta.dio_device.d_config_bit(
                self._meta.uldaq.DigitalPortType.AUXPORT,
                mock_channel,
                item["direction"],
            )
            self._meta.instr._check_is_open.assert_called_once_with()

    def test_get_dio_input_bit(self):
        """MCC_USB1808X.get_dio_input_bit(), happy flow."""
        self._meta.instr._check_is_open = MagicMock()
        device_meta = self._mock_device()
        self._meta.instr.open()
        mock_channel = MagicMock()
        rt_val = self._meta.instr.get_dio_input_bit(mock_channel)
        self.assertEqual(
            rt_val,
            device_meta.dio_device.d_bit_in(
                self._meta.uldaq.DigitalPortType.AUXPORT, mock_channel
            ),
        )
        self._meta.instr._check_is_open.assert_called_once_with()

    def test_set_dio_output_bit(self):
        """MCC_USB1808X.set_dio_output_bit(), happy flow."""
        self._meta.instr._check_is_open = MagicMock()
        device_meta = self._mock_device()
        self._meta.instr.open()
        mock_channel = MagicMock()
        mock_value = MagicMock()
        self._meta.instr.set_dio_output_bit(mock_channel, mock_value)
        device_meta.dio_device.d_bit_out.assert_called_once_with(
            self._meta.uldaq.DigitalPortType.AUXPORT, mock_channel, mock_value
        )
        self._meta.instr._check_is_open.assert_called_once_with()

    def test_get_ai_num_channels(self):
        """MCC_USB1808X.get_ai_num_channels(), happy flow."""
        self._meta.instr._check_is_open = MagicMock()
        device_meta = self._mock_device()
        self._meta.instr.open()
        rt_val = self._meta.instr.get_ai_num_channels()
        device_meta.ai_device.get_info.assert_called_once_with()
        device_meta.ai_device.get_info().get_num_chans.assert_called_once_with()
        self.assertEqual(rt_val, device_meta.ai_device.get_info().get_num_chans())
        self._meta.instr._check_is_open.assert_called_once_with()

    def test_get_ai_ranges(self):
        """MCC_USB1808X.get_ai_ranges(), happy flow."""
        self._meta.instr._check_is_open = MagicMock()
        device_meta = self._mock_device()
        self._meta.instr.open()
        mock_resp = [MagicMock(), MagicMock(), MagicMock()]
        device_meta.ai_device.get_info().get_ranges = MagicMock(return_value=mock_resp)
        device_meta.ai_device.get_info.reset_mock()
        rt_val = self._meta.instr.get_ai_ranges()
        device_meta.ai_device.get_info.assert_called_once_with()
        device_meta.ai_device.get_info().get_ranges.assert_called_once_with(
            self._meta.uldaq.AiChanType.VOLTAGE
        )
        self.assertEqual(rt_val[0], mock_resp[0].name)
        self.assertEqual(rt_val[1], mock_resp[1].name)
        self.assertEqual(rt_val[2], mock_resp[2].name)
        self._meta.instr._check_is_open.assert_called_once_with()

    def test_get_ai_value(self):
        """MCC_USB1808X.get_ai_value(), happy flow."""
        self._meta.instr._check_is_open = MagicMock()
        device_meta = self._mock_device()
        self._meta.instr.open()
        mock_channel = MagicMock()
        mock_input_mode = MagicMock()
        mock_analog_range = MagicMock()
        rt_val = self._meta.instr.get_ai_value(
            mock_channel, mock_input_mode, mock_analog_range
        )
        self.assertEqual(
            rt_val,
            device_meta.ai_device.a_in(
                mock_channel,
                self._meta.uldaq.AiInputMode[mock_input_mode],
                self._meta.uldaq.Range[mock_analog_range],
                self._meta.uldaq.AInFlag.DEFAULT,
            ),
        )
        self._meta.instr._check_is_open.assert_called_once_with()

    def test_get_ao_num_channels(self):
        """MCC_USB1808X.get_ao_num_channels(), happy flow."""
        self._meta.instr._check_is_open = MagicMock()
        device_meta = self._mock_device()
        self._meta.instr.open()
        rt_val = self._meta.instr.get_ao_num_channels()
        device_meta.ao_device.get_info.assert_called_once_with()
        device_meta.ao_device.get_info().get_num_chans.assert_called_once_with()
        self.assertEqual(rt_val, device_meta.ao_device.get_info().get_num_chans())
        self._meta.instr._check_is_open.assert_called_once_with()

    def test_get_ao_ranges(self):
        """MCC_USB1808X.get_ao_ranges(), happy flow."""
        self._meta.instr._check_is_open = MagicMock()
        device_meta = self._mock_device()
        self._meta.instr.open()
        mock_resp = [MagicMock(), MagicMock(), MagicMock()]
        device_meta.ao_device.get_info().get_ranges = MagicMock(return_value=mock_resp)
        device_meta.ao_device.get_info.reset_mock()
        rt_val = self._meta.instr.get_ao_ranges()
        device_meta.ao_device.get_info.assert_called_once_with()
        device_meta.ao_device.get_info().get_ranges.assert_called_once_with()
        self.assertEqual(rt_val[0], mock_resp[0].name)
        self.assertEqual(rt_val[1], mock_resp[1].name)
        self.assertEqual(rt_val[2], mock_resp[2].name)
        self._meta.instr._check_is_open.assert_called_once_with()

    def test_set_ao_value(self):
        """MCC_USB1808X.set_ao_value(), happy flow."""
        self._meta.instr._check_is_open = MagicMock()
        device_meta = self._mock_device()
        self._meta.instr.open()
        mock_channel = MagicMock()
        mock_analog_range = MagicMock()
        mock_value = MagicMock()
        self._meta.instr.set_ao_value(mock_channel, mock_analog_range, mock_value)
        device_meta.ao_device.a_out.assert_called_once_with(
            mock_channel,
            self._meta.uldaq.Range[mock_analog_range],
            self._meta.uldaq.AOutFlag.DEFAULT,
            mock_value,
        )
        self._meta.instr._check_is_open.assert_called_once_with()



"""Convenience functions to handle the singleton-instance of QMI_Context.

Most QMI applications will create a single instance of `QMI_Context` and
use that context for all interactions with QMI. This module provides
functions to make this way of working more convenient.

The public functions defined in this module, are also imported into
the top-level `qmi` package. Applications can therefore call these
functions directly, for example::

    import qmi
    qmi.start("my_context")
    proxy = qmi.make_instrument("my_instrument", InstrumentClass, ...)
    qmi.stop()
"""

import sys
import logging
import os
import os.path
import time

from typing import Any, Optional, Type, Dict, List, cast

import qmi.core.config
import qmi.core.logging_init
import qmi.core.thread

from qmi.core.config_struct import config_struct_from_dict
from qmi.core.config_defs import CfgQmi, CfgContext
from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_UsageException, QMI_NoActiveContextException, QMI_ConfigurationException
from qmi.core.instrument import QMI_Instrument
from qmi.core.rpc import QMI_RpcObject
from qmi.core.task import QMI_Task, QMI_TaskRunner
from qmi.core.util import format_address_and_port


# Global variable holding the current QMI_Context instance.
_qmi_context = None

# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)

# global constant holding environment variables
QMI_CONFIG = os.getenv("QMI_CONFIG")


def context() -> QMI_Context:
    """Return the current QMI_Context instance.

    This function may only be called when there is an active QMI context
    (i.e. after calling `qmi.start()`).

    Raises:
        QMI_NoActiveContextException: If there is no active context.
    """
    if _qmi_context is None:
        raise QMI_NoActiveContextException("No active context")
    return _qmi_context


def start(
        context_name: str,
        config_file: Optional[str] = None,
        init_logging: bool = True,
        console_loglevel: Optional[str] = None,
        context_cfg: Optional[dict] = None) -> None:
    """Create and start a global QMI_Context instance.

    This function should be called exactly once by the top-level code of
    the application.

    This function performs a number of initialization steps:

    * read the QMI configuration file;
    * initialize logging;
    * create and start a global QMI context.

    The context name is a short string, without spaces or strange characters.
    This name will be used to refer to this context from other processes,
    as well as for displaying and logging.

    The context name should be unique within the workgroup (i.e. among all
    Python processes that work together via QMI.) If this context acts as
    a server for other contexts to connect to, its name *must* be unique.
    Otherwise, if this context only acts as a client, a duplicate context name
    is allowed but not recommended.

    If `config_file` is provided, it should be a path to a valid QMI
    configuration file. The referenced file will be used to load the
    configuration from.
    If `config_file` is ``None`` and the environment variable `QMI_CONFIG` is
    set, the configuration will be loaded from the file that the environment
    variable points to. If `QMI_CONFIG` is not set, an empty configuration is
    used.

    The configuration created can be edited by providing a dictionary as an
    optional input. See config_defs.CfgContext class for possible dict keys
    and respective values and types that are allowed. Wrong (type) of keys
    and/or values will raise an exception.

    Parameters:
        context_name: Name of the QMI context.
        config_file: Optional path to the QMI configuration file.
        init_logging: Optional flag; set False to skip logging initialization.
        console_loglevel: Optionally override `console_loglevel` from config file.
        context_cfg: Optionally insert or override context(s) in config.contexts
    """

    global _qmi_context

    qmi.core.thread.check_in_main_thread()

    if _qmi_context is not None:
        raise QMI_UsageException("QMI context already started")

    config = create_config_from_file(config_file)  # this will raise FileNotFoundError if config_file == ""

    # Check for insert or override of context or contexts.
    if context_cfg is not None:
        for key in context_cfg.keys():
            config.contexts.update({key: config_struct_from_dict(context_cfg[key], CfgContext)})

    # Check for loglevel instructions
    if console_loglevel is not None:
        valid_levels = ["INFO", "WARNING", "DEBUG", "CRITICAL", "FATAL", "ERROR", "WARN", "NOTSET"]
        if console_loglevel not in valid_levels:
            raise QMI_ConfigurationException(f"Trying to use non-valid console loglevel {console_loglevel}. " +
                                             f"Valid values are {valid_levels}.")

        config.logging.console_loglevel = console_loglevel.upper()

    # Create QMI context.
    _qmi_context = QMI_Context(context_name, config)

    if init_logging:
        _init_logging()

    _logger.info("QMI starting (prog=%r, context=%r, pid=%d, config=%r)",
                 sys.argv[0], context_name, os.getpid(), config_file)

    # Start QMI context.
    _qmi_context.start()

    # Connect to peer contexts.
    _connect_to_peers()


def create_config_from_file(config_file: Optional[str]) -> CfgQmi:
    """Create configuration from a file.

    Parameters:
        config_file: Path of configuration file or None.

    Returns:
        Top-level QMI configuration structure.
    """
    # Try first to see if input is given or QMI_CONFIG is set
    config_file = config_file if config_file is not None else QMI_CONFIG
    if config_file is None:
        # Failing that, use the default class
        return CfgQmi()

    else:
        _logger.debug("Reading QMI configuration file %s", config_file)
        cfgdict = qmi.core.config.load_config_file(config_file)
        cfgdict["config_file"] = os.path.abspath(config_file)

    return config_struct_from_dict(cfgdict, CfgQmi)


def _init_logging() -> None:
    """Initialize Python logging framework."""

    # Get logging configuration.
    assert _qmi_context is not None
    cfg = _qmi_context.get_config()
    loglevel = cfg.logging.loglevel
    console_loglevel = cfg.logging.console_loglevel
    logfile = cfg.logging.logfile
    loglevels = cfg.logging.loglevels

    # If QMI_DEBUG is set, override configured log level.
    if os.getenv("QMI_DEBUG"):
        loglevel = "DEBUG"
        console_loglevel = "DEBUG"

    if logfile:
        # Expand placeholders in logfile name.
        # For example "log_%(context)s_%(datetime)s.txt"
        gmtime = time.gmtime()
        attrs = {
            "context": _qmi_context.name,
            "date": time.strftime('%Y-%m-%d', gmtime),
            "datetime": time.strftime('%Y-%m-%dT%H-%M-%S', gmtime),
            }
        logfile = logfile % attrs

        # Make absolute file name.
        qmi_home = _qmi_context.get_qmi_home_dir()
        logfile = os.path.join(qmi_home, logfile)

    # Initialize logging.
    qmi.core.logging_init.start_logging(loglevel=loglevel,
                                        console_loglevel=console_loglevel,
                                        logfile=logfile,
                                        loglevels=loglevels,
                                        rate_limit=cfg.logging.rate_limit,
                                        burst_limit=cfg.logging.burst_limit)


def _connect_to_peers() -> None:
    """Helper function to connect to the pre-configured list of peers."""

    assert _qmi_context is not None

    # Get list of peer contexts.
    cfg = _qmi_context.get_config()
    ctxcfg = _qmi_context.get_context_config()

    # For each peer.
    for peername in ctxcfg.connect_to_peers:

        # Get host and port number of peer context.
        peercfg = cfg.contexts.get(peername)
        if not peercfg:
            raise QMI_ConfigurationException("Can not connect to unknown peer {!r}".format(peername))
        peer_host = peercfg.host
        peer_port = peercfg.tcp_server_port
        if (not peer_host) or (not peer_port):
            raise QMI_ConfigurationException("Missing host/port for peer context {!r}".format(peername))

        # Connect to peer.
        peer_address = format_address_and_port((peer_host, peer_port))
        _logger.info("Connecting to peer context %r on %s", peername, peer_address)
        _qmi_context.connect_to_peer(peername, peer_address)


def stop() -> None:
    """Stop and destroy the `QMI_Context` instance.

    This function should be called exactly once when the application exits.

    Raises:
        QMI_NoActiveContextException: If there is no active context.
    """

    global _qmi_context

    qmi.core.thread.check_in_main_thread()

    if _qmi_context is None:
        raise QMI_NoActiveContextException()

    _qmi_context.stop()
    _qmi_context = None


def info() -> str:
    """Return human-readable information about the context."""

    if _qmi_context is None:
        return "*** No active QMI context ***"
    else:
        return _qmi_context.info()


def make_rpc_object(rpc_object_name: str,
                    rpc_object_class: Type[QMI_RpcObject],
                    *args: Any,
                    **kwargs: Any
                    ) -> Any:
    """Create an instance of a `QMI_RpcObject` subclass and make it accessible via RPC.

    The actual object instance will be created in a separate background thread.
    To access the object, you can call its methods via RPC. Note that using the returned proxy in
    a thread might require casting it first: proxy = cast(proxy, rpc_object_class)(qmi.context(), rpc_object_name).

    Parameters:
        object_name: Unique name for the new object instance.
            This name will also be used to access the object via RPC.
        object_class: Class that implements this object (must be a subclass of `QMI_RpcObject`).
        args: Optional arguments for the object class constructor.

    Returns:
        An RPC proxy that provides access to the new object instance.
    """
    if _qmi_context is None:
        raise QMI_NoActiveContextException()

    return _qmi_context.make_rpc_object(rpc_object_name, rpc_object_class, *args, **kwargs)


def make_instrument(instrument_name: str,
                    instrument_class: Type[QMI_Instrument],
                    *args: Any,
                    **kwargs: Any
                    ) -> Any:
    """Create an instance of a `QMI_Instrument` subclass and make it accessible via RPC.

    The actual instrument instance will be created in a separate background thread.
    To access the instrument, you can call its methods via RPC.

    Parameters:
        instrument_name: Unique name for the new instrument instance.
            This name will also be used to access the instrument via RPC.
        instrument_class: Class that implements this instrument (must be a subclass of `QMI_Instrument`).
        args: Optional arguments for the instrument class constructor.

    Returns:
        An RPC proxy that provides access to the new instrument instance.
    """
    if _qmi_context is None:
        raise QMI_NoActiveContextException()

    return _qmi_context.make_instrument(instrument_name, instrument_class, *args, **kwargs)


def make_task(task_name: str,
              task_class: Type[QMI_Task],
              *args: Any,
              task_runner: Type[QMI_TaskRunner] = QMI_TaskRunner,
              **kwargs: Any
              ) -> Any:
    """Create an instance of a `QMI_Task` subclass and make it accessible via RPC.

    The actual task instance will be created in a separate thread.
    A `QMI_TaskRunner` will be created to manage the task thread.
    The task can be accessed by calling the methods of the task runner via RPC.

    Note that the task is not yet started.
    To start the task, perform an explicit call to the `start()` method of the returned task.

    Parameters:
        task_name: Unique name for the new task instance.
            This name will also be used to access the task runner via RPC.
        task_class: Class that implements this task (must be a subclass of `QMI_Task`).
        task_runner: Class that implements the managing of the task (must be a subclass of `QMI_Taskrunner`)
        args: Optional arguments for the task class constructor.

    Returns:
        An RPC proxy that provides access to the new task.
    """
    if _qmi_context is None:
        raise QMI_NoActiveContextException()

    return _qmi_context.make_task(task_name, task_class, *args, task_runner=task_runner, **kwargs)


def list_rpc_objects(rpc_object_baseclass=None) -> List:
    """Show a list of RPC objects in the local context and peer contexts."""

    if _qmi_context is None:
        raise QMI_NoActiveContextException()

    return _qmi_context.list_rpc_objects(rpc_object_baseclass)


def show_rpc_objects(rpc_object_baseclass=None) -> None:
    """Show a list of RPC objects in the local context and peer contexts."""

    if _qmi_context is None:
        raise QMI_NoActiveContextException()

    return _qmi_context.show_rpc_objects(rpc_object_baseclass)


def show_instruments() -> None:
    """Show a list of instruments in the local context and peer contexts."""
    if _qmi_context is None:
        raise QMI_NoActiveContextException()

    return _qmi_context.show_instruments()


def show_tasks() -> None:
    """Show a list of tasks in the local context and peer contexts."""
    if _qmi_context is None:
        raise QMI_NoActiveContextException()

    return _qmi_context.show_tasks()


def show_contexts() -> None:
    """Show a list of currently connected contexts (including the local context)."""
    if _qmi_context is None:
        raise QMI_NoActiveContextException()

    return _qmi_context.show_contexts()


def show_network_contexts() -> None:
    """Show a lists of context on the network (connected and not connected)."""
    if _qmi_context is None:
        raise QMI_NoActiveContextException()

    return _qmi_context.show_network_contexts()


def get_rpc_object(rpc_object_name: str, auto_connect: bool = False, host_port: Optional[str] = None) -> Any:
    """Return a proxy for the specified object.

    The object may exist either in the local context, or in a peer context.

    Parameters:
        object_name: Object name, formatted as ``"<context_name>.<object_name>"``.
        auto_connect: if True, connect automatically to the object peer.
        host_port: Optional host:port string pattern to guide the auto_connect.

    Returns:
        A proxy for the specified object.
    """
    if _qmi_context is None:
        raise QMI_NoActiveContextException()

    return _qmi_context.get_rpc_object_by_name(rpc_object_name, auto_connect=auto_connect, host_port=host_port)


def get_instrument(instrument_name: str, auto_connect: bool = False, host_port: Optional[str] = None) -> Any:
    """Return a proxy for the specified instrument.

    The instrument may exist either in the local context, or in a peer context.

    Parameters:
        instrument_name: Instrument name, formatted as ``"<context_name>.<instrument_name>"``.
        auto_connect: if True, connect automatically to the instrument peer.
        host_port: Optional host:port string pattern to guide the auto_connect.

    Returns:
        A proxy for the specified instrument.
    """
    if _qmi_context is None:
        raise QMI_NoActiveContextException()

    return _qmi_context.get_instrument(instrument_name, auto_connect, host_port)


def get_task(task_name: str, auto_connect: bool = False, host_port: Optional[str] = None) -> Any:
    """Return a proxy for the specified task.

    The task may exist either in the local context, or in a peer context.

    Parameters:
        task_name: Task name, formatted as ``"<context_name>.<task_name>"``.
        auto_connect: if True, connect automatically to the task peer.
        host_port: Optional host:port string pattern to guide the auto_connect.

    Returns:
        A proxy for the specified task.
    """
    if _qmi_context is None:
        raise QMI_NoActiveContextException()

    return _qmi_context.get_task(task_name, auto_connect, host_port)


def get_configured_contexts() -> Dict[str, CfgContext]:
    """Return a dictionary of active QMI contexts.

    Returns:
        An OrderedDict object of all active QMI contexts.
    """
    if _qmi_context is None:
        raise QMI_NoActiveContextException()

    return _qmi_context.get_configured_contexts()



"""
Wavelength Electronics, temperature controller.

The qmi.instruments.wavelength package provides support for:
- TC LAB series.
"""
from qmi.instruments.wavelength.tclab import AutotuneMode, TemperatureControllerCondition
# Alternative, QMI naming convention approved name
from qmi.instruments.wavelength.tclab import Wavelength_TC_Lab as Wavelength_TcLab
